//****************************************************************************
// Verilog version of Adept VHDL 1394 interface.
//
// (C) 2013, Ken Chamberlain <kenneth.r.chamberlain@comcast.net>
//
// Feel free to use this code in any project (commercial or not), as long as you
// keep this message, and the copyright notice. This code is provided "as is", 
// without any warranties of any kind. 
// 
//****************************************************************************

//`define debug
//`define la

module IOExpansion(clk, EppAstb, EppDstb, EppWr, EppDB, EppWait, Led,
                   //regEppAdrw, cpu_reg,
                   dma_a, dma_dw, dma_dr, dma_ack, data_ready,
                   dma_req, dma_rw, dma_hiram, dma_loram, dma_charen, dma_GAME, dma_EXROM,
                   cpu_reset, cpu_pause, cpu_ack
`ifdef la
                   ,la_clk, clear_trigger, la_offset
`endif
`ifdef debug
                   ,breakTrig, break_pc
`endif
                   );

// original stuff
input EppAstb;     // Address strobe
input EppDstb;     // Data strobe
input EppWr;       // Port write signal
input [7:0] Led;   // 8 non-virtual LEDs on the ECB
inout [7:0] EppDB; // Port data bus
output EppWait;    // Port wait signal

//my stuff
input clk;
input [7:0] dma_dr;
//input [7:0] cpu_reg;
input dma_ack, data_ready, cpu_ack;

`ifdef la
input la_clk;
output la_offset;
output clear_trigger;
`endif

`ifdef debug
input breakTrig;
output [15:0] break_pc;
`endif

//output [5:0] regEppAdrw;
output [7:0] dma_dw;
output [15:0] dma_a;
output dma_req, dma_rw, dma_hiram, dma_loram, dma_charen, dma_GAME, dma_EXROM;
output cpu_reset;
output cpu_pause;

parameter H = 1'b1;
parameter L = 1'b0;

//----------------------------------------------------------------------------------
//-- Company: Digilent RO
//-- Engineer: Mircea Dabacan
//-- 
//-- Create Date:    20:46:18 04/06/2009 
//-- Design Name: Adept IO Expansion Reference Component
//-- Module Name:    IOExpansion - Behavioral 
//-- Project Name: Adept 2
//-- Target Devices: 
//-- Tool versions: WP10.1.03
//-- Description: 
//--
//-- The component implements EPP byte registers for virtual IOs on a FPGA target: 
//
//--   Name              Epp     Dir Explain
//--                     Address 
//-- regVer(7 downto 0)  0x00    I/O returns the complement of written value
//-- Led(7 downto 0)     0x01    In  8 virtual LEDs on the PC I/O Ex GUI
//************************************************************
//my stuff
//-- dma_a[7:0]      0x20       In     lower dma address bits
//-- dma_a[15:8]     0x21       In     upper dma address bits
//-- dma_data[7:0]   0x22       In/Out dma data r/w bits
//-- dma_ctrl[7:0]   0x23       In     control bits - dma memory space, cpu control and logic analyzer enable
//-- la_arm          0x24       In     arm with any write if enabled
//-- cpu_single_step 0x25       In     1 clock cycle with any write
//-- debug_display   0x26       In     debug data display select (4) - toggle with any write
//-- dma_data2[7:0]  0x28       In/Out dma data r/w bits with address increment
//-- cpu_reg[7:0]    0b11xxxxxx Out    internal cpu register read access
//-- break_pc[7:0]   0x29       In     lower pc break address bits
//-- break_pc[15:8]  0x2A       In     upper pc break address bits

`ifdef debug
reg [15:0] break_pc = 16'h01FF;
`endif

`ifdef la
reg [7:0] la_offset;
reg clear_trigger = H;
reg trigger_clear2 = H;
reg trigger_clear = H;
`endif

reg [15:0] dma_a;
reg [7:0] regEppAdr, regVer, dma_dw, dma_ctrl = 8'hFF, dma_ctrl2;
reg [1:0] state = 2'b00;
reg EppWait1, EppWait2, EppWait3;
reg dma_read_cycle, dma_read_cycle2, dma_read_cycle3, dma_read_cycle4, dma_read_pend; 
reg dma_write_cycle, dma_write_cycle2, dma_write_cycle3, dma_write_cycle4, dma_write_pend; 
reg reg_write_cycle, reg_write_cycle2, reg_write_cycle3; 
reg dma_req = H;
reg dma_rw = H;
reg single_step = H;
reg single_step2 = H;
reg ss = L;
reg ss_pending = L;
reg inc_wr_addr = L;

wire [7:0] busEppInternal;
wire #10 EppDstbDelayed = EppDstb;
//wire [5:0] regEppAdrw = regEppAdr[5:0];


//-- Epp signals
//   -- Port signals
// the state machine now drives this signal to allow for dma delays...
//assign EppWait = ~EppAstb | ~EppDstb; //original
reg EppWait;

//  -- EPP Address/Data Read
assign EppDB = EppWr ? busEppInternal : 8'hz;
assign busEppInternal = ~EppAstb ? regEppAdr :
                        (regEppAdr == 8'h00) ? regVer : //required register for adept program to communicate
                        (regEppAdr == 8'h01) ? Led :
                        (regEppAdr == 8'h20) ? dma_a[7:0]    :
                        (regEppAdr == 8'h21) ? dma_a[15:8]   :
                        (regEppAdr == 8'h22) ? dma_dr[7:0]   :
                        (regEppAdr == 8'h28) ? dma_dr[7:0]   :
`ifdef debug
                        (regEppAdr == 8'h23) ? dma_ctrl[7:0] : 
                        (regEppAdr == 8'h29) ? break_pc[7:0]  :
                        (regEppAdr == 8'h2A) ? break_pc[15:8] : 8'hEA;
`else
                        (regEppAdr == 8'h23) ? dma_ctrl[7:0] : 8'hEA; 
`endif
                         //(regEppAdr[7] | regEppAdr[6]) ? cpu_reg[7:0] : 8'hEA;


//  -- EPP Address Write register
always @(posedge EppAstb)
  begin
    if(~EppWr) regEppAdr <= EppDB;
  end

//  -- EPP Data Read - Check for DMA
always @(negedge EppDstb)
  begin
    if(EppDstbDelayed & EppWr & ((regEppAdr == 8'h22) | (regEppAdr == 8'h28)))
      begin
        //toggling the state of this line triggers a dma read cycle
        dma_read_cycle <= ~dma_read_cycle;
      end
    else dma_read_cycle <= dma_read_cycle;
  end
  
//  -- EPP Data Write
always @(posedge EppDstb)
  begin
    if(~EppWr)
      begin
        if((regEppAdr == 8'h22) | (regEppAdr == 8'h28))
          begin
            dma_dw[7:0] <= EppDB;   //dma data write occuring
            //toggling the state of this line triggers a dma write cycle
            dma_write_cycle <= ~dma_write_cycle;
            reg_write_cycle <= reg_write_cycle;
            //for block write - pre increment!
            if((regEppAdr == 8'h28) & inc_wr_addr) dma_a <= dma_a + H;
          end
        else
          begin
            dma_write_cycle <= dma_write_cycle;
            reg_write_cycle <= ~reg_write_cycle;
            
            if(regEppAdr == 8'h00)            // required register for adept program to communicate
              regVer <= 8'hFF ^ EppDB;        // register update (complemented)
            else if(regEppAdr == 8'h20)
              dma_a[7:0] <= EppDB;            // dma address low
            else if(regEppAdr == 8'h21)
              dma_a[15:8] <= EppDB;           // dma address high
            else if(regEppAdr == 8'h23)
              dma_ctrl2[7:0] <= EppDB[7:0];     // dma control
`ifdef la
            else if(regEppAdr == 8'h24)
              begin
                trigger_clear <= ~trigger_clear; // clear logic analyzer trigger - arm it
                la_offset <= EppDB; //load trigger display offset - 0 is closer to the right edge of the screen
              end                   //                              255 is closer to the left edge of the screen
`endif
            else if(regEppAdr == 8'h25)
              single_step <= ~single_step;     // single step the cpu
`ifdef debug
            else if(regEppAdr == 8'h29)
              break_pc[7:0] <= EppDB;           // pc break address low
            else if(regEppAdr == 8'h2A)
              break_pc[15:8] <= EppDB;          // pc break address high
`endif
          end
      end
    else
      begin
        // read cycle!
        dma_write_cycle <= dma_write_cycle;
        reg_write_cycle <= reg_write_cycle;
        //for block read - post increment!
        if(regEppAdr == 8'h28) dma_a <= dma_a + H;
      end
  end

`ifdef la
always @(posedge la_clk)
  begin
    trigger_clear2 <= trigger_clear;
    clear_trigger <= ~(trigger_clear2 ^ trigger_clear);
  end
`endif
  
//procesing for dma and other stuff
always @(posedge clk)
  begin
    single_step2 <= single_step;
    ss <= single_step2 ^ single_step;
    
    ss_pending <= ss & cpu_ack | ss_pending & cpu_ack;
    
    //synchonize and delay the wait signal in case a dma cycle is initiated
    EppWait1 <= ~EppAstb | ~EppDstb;
    EppWait2 <= EppWait1;
    EppWait3 <= EppWait2;
    
    //synchronize end of a datastrobe dma write
    dma_write_cycle2 <= dma_write_cycle;
    dma_write_cycle3 <= dma_write_cycle2;
    dma_write_cycle4 <= dma_write_cycle3;

    //synchronize beginning of a datastrobe dma read
    dma_read_cycle2 <= dma_read_cycle;
    dma_read_cycle3 <= dma_read_cycle2;
    dma_read_cycle4 <= dma_read_cycle3;
    
    //dma initiated, these will only stay active for 1 cycle
    dma_read_pend <= dma_read_cycle4 ^ dma_read_cycle3;
    dma_write_pend <= dma_write_cycle4 ^ dma_write_cycle3;

    reg_write_cycle2 <= reg_write_cycle;
    reg_write_cycle3 <= reg_write_cycle2;
    
    if(reg_write_cycle2 ^ reg_write_cycle3)
      begin
        if(regEppAdr == 8'h23) dma_ctrl <= dma_ctrl2;
      end
`ifdef debug
    else if(breakTrig)
      begin
        dma_ctrl[6] <= L;
      end
`endif
    
    // dma cycle state machine / pport wait interface
    case(state)
      2'h0:
        begin
          if(dma_read_pend)
            begin
              inc_wr_addr <= L;
              state <= 4'h1;
              dma_req <= L;
              EppWait <= L; // keep it low
              dma_rw <= H;
            end
          else if(dma_write_pend)
            begin
              inc_wr_addr <= L;
              state <= 4'h2;
              dma_req <= L;
              EppWait <= H; // keep it high
              dma_rw <= L;
            end
          else
            begin
              state <= 4'h0;
              dma_req <= H;
              EppWait <= EppWait3; // let it be whatever it is
              dma_rw <= H;
            end
        end
      
      2'h1:
        begin
          if(~data_ready)
            begin
              state <= 4'h0;
              dma_req <= H;
              EppWait <= EppWait3; // should be low by now
              dma_rw <= H;
            end
          else
            begin
              if(~dma_ack)
                dma_req <= H;
              state <= 4'h1;
              EppWait <= L; // delay this transaction
              dma_rw <= H;
            end
        end

      2'h2:
        begin
          if(~data_ready)
            begin
              state <= 4'h0;
              dma_req <= H;
              EppWait <= EppWait3; // should be low by now
              dma_rw <= H;
              inc_wr_addr <= regEppAdr == 8'h28;
            end
          else
            begin
              if(~dma_ack)
                begin
                  dma_req <= H;
                  dma_rw <= H;
                end
              state <= 4'h2;
              EppWait <= H; // delay next transaction
            end
        end

      //should never get here
      default :
        begin
          state <= 4'h0;
          dma_req <= H;
          EppWait <= EppWait3;
          dma_rw <= H;
        end
    endcase
  end
  
//control register signals
assign dma_loram = dma_ctrl[0];
assign dma_hiram = dma_ctrl[1];
assign dma_charen = dma_ctrl[2];
assign dma_GAME = dma_ctrl[3];
assign dma_EXROM = dma_ctrl[4];
assign cpu_reset = dma_ctrl[5];
assign cpu_pause = dma_ctrl[6] | ((ss | ss_pending) & cpu_ack);
//assign ? = dma_ctrl[7];

endmodule
