//****************************************************************************
// 6510/6502 verilog cpu. Von Nuemann architecture.
//
// (C) 2013, Ken Chamberlain <kenneth.r.chamberlain@comcast.net>
//
// Feel free to use this code in any project (commercial or not), as long as you
// keep this message, and the copyright notice. This code is provided "as is", 
// without any warranties of any kind. 
// 
//****************************************************************************

//****************************************************************************
//X6510

//Read data is not available for use until the next state after the state the
//read address was put onto the cpu_a bus because of memory pipeline delay.

//The first read port is mainly used for reading instructions or addresses.
//This is a read only port.

//The second port is used for all operand and stack byte r/w operations.
//The external memory must be dual ported so simultaneous memory acceses can occur.

//Most instructions take fewer clock cycles than the original processor.
//mos6502 clock cycle counts, X6510 cycle counts are noted in the iDecode file.

//CPU states (S_EXECUTE_INSTn) are the starting instruction execute states
//for all instructions based on it were a 1, 2 or 3 byte instruction.
// Some previous state was used to load the instruction into the ir queue.
//To calculate cycle counts start counting at state S_EXECUTE_INSTn until you
//get back to state S_EXECUTE_INSTn.  Do not count state S_EXECUTE_INSTn again
//as that is now processing the next instruction.

//Unimplemented instructions will halt the processor.

//BCD arithmatic is supported but takes 1 additional clock cycle to complete.

//The processor can be stopped at any time with a high level on pause or
//low levels on vic_req or dma_req.

//All instruction fetches are initiated from states 2-7 and 11.
//If the instruction cache cannot take in 4 new bytes the next instruction
//register fetch address is still read but the data is disregarded until
//there is room for it in the 12 byte instruction register.  At least 12 bytes
//are required in the instruction queue.  If instruction sequences are encounted
//that require more the queue length will have to be increased.

//Two types of dma are supported, burst type which is used by VIC requests
//for video and single cycle requests that can be used by some other device.
//Burst requests are acknowledged by vic_ack going low and will always occur
//on the next memory cycle of the request. A single cycle dma request is
//acknowleged by dma_ack. This is a lower priority dma and may be held off
//by a burst request at the same time or already in progress even though
//each use a different port for accessing memory.  The burst requests never
//write to memory.

//while cpu_ack2 is low and cpu_ack is high the contents of the data read
//buses are saved off to a register for use on the first cpu_ack cycle when
//bus mastership has been regained. There are no memory cycles lost during
//bus mastership changes.

//`define N6502    //remove the C64 memory map register...
//`define pport    //allow registers to be accessed externally...
//`define debug      //video display of registers, pc breakpoint & single stepping.
`define zps_w_thru //allow zp and stack writes to external ram so code and jump vectors can exist
                   //in this memory space - slows ZP write instructions by 1 cycle.
//              __    __    __    __    __    _     __    __    __    __    __    __    __    __
//     clk  |__|  |__|  |__|  |__|  |__|  |__| ..._|  |__|  |__|  |__|  |__|  |__|  |__|  |__|
//
//   cpu_a(l)  X A0  X A1  X A2  X A3  X V1  X ... X Vm  X Vn  X XX  X A4  X A5  X A6  X A7  X
//
//  cpu_ib     X XX  X i0  X i1  X i2  X i3  X ... X XX  X XX  X XX  X XX  X i4  X i5  X i6  X
//
//  cpu_aw     X B0  X B1  X B2  X B3  X XX  X ... X XX  X XX  X D0  X B4  X B5  X B6  X B7  X
//             _______       __________________   ______________       _________________________
//  cpu_rw            |_____|                  ...              |_____|
//
//  cpu_db     X XX  X O0  X XX  X O2  X O3  X ... X XX  X XX  X XX  X XX  X O4  X O5  X O6  X
//
//  cpu_wb     X XX  X w1  X XX  X XX  X XX  X ... X XX  X XX  X wd  X XX  X XX  X XX  X XX  X
//             ____________________                         ____________________________________
// vic_req*                        |___________..._________|
//             _________________________                         _______________________________
// vic_ack*                             |______...______________|
//             ____________________________                            _________________________
// dma_req*                                |___...____________________|
//             ________________________________   ______________       _________________________
// dma_ack*                                    ...              |_____|
//                                       ______   ____________________
// cpu_ack*  XX_________________________|      ...                    |_________________________
//                                                __________________________
// cpu_ack2* XXXXXXXXX________________________|...                          |___________________
//              __    __    __    __    __    _     __    __    __    __    __    __    __    __
//     clk  |__|  |__|  |__|  |__|  |__|  |__| ..._|  |__|  |__|  |__|  |__|  |__|  |__|  |__|


module X6510V(clk,      // system memory clock
              cpu_ack,  // this is low when the cpu has control of the memory bus
              cpu_a,    // [15:0] cpu/vic address for accessing cpu_ib[31:16] data (Read-Only)
              cpu_al,   // [15:2] cpu/vic address for accessing cpu_ib[15:0] data (Read-Only)
              cpu_ib,   // [31:0] instrution, address or memory operand fetching data bus
              pRead,    // indicates the access was a peripheral and should be read off of the cpu_db bus
              cpu_db,   // [7:0] data peripheral operand read data
              cpu_wb,   // [7:0] cpu data operand write data
              cpu_rw,   // read(H) / write(L) control signal
              cpu_wbyte,// [3:0] write{L} byte selects (may not be needed depending on memory config)
              vic_req,  // burst read dma request
              vic_ack,  // burst read dma acknowledge - 1 cycle delay from request
              vic_a,    // [15:1] burst read source address => cpu_a, cpu_al
              dma_req,  // single byte dma access request *must be released on the next clk dma_ack is asserted
              dma_ack,  // single byte dma acknowledge - next cycle free from burst request
              dma_a,    // [15:0] single byte source address => cpu_aw
              dma_dw,   // [7:0] single byte dma data write source => cpu_wb
              dma_rw,   // single byte dma read(H) / write(L) control signal => cpu_rw
              nmi_req,  // negative edge non-maskable interrupt input
              irq,      // low level maskable interrupt input
              resetIn,  // low level cpu reset input
              cpu_pause // low level cpu pause input
             
`ifndef N6502
              , cpu_hiram  // C64 memory map control signal
              , cpu_loram  // C64 memory map control signal
              , cpu_charen // C64 memory map control signal
`endif             
`ifdef debug
              , debug_add // [4:0] video select of what to be displayed on its current raster
              , debug_Dout// [47:0] 8 x 6 bits data value to be displayed
              , debug_sel // [2:0] selects what dataset is to be displayed, reigisters, pc trace etc...
              , breakTrig // low level output that indicates the pc breakpoint was reached (should cause pause to occur)
              , break_pc  // [15:0] pc compare address for program breakpoint
`endif
`ifdef pport
              , pp_add    // select address of an internal register
              , pp_Dout   // [7:0] internal register value or partial value
              //,EppDstb  // can be used to trigger and internal address increment for dumping out a trace buffer
`endif
             );

input clk;

input vic_req;
input [15:1] vic_a;

input nmi_req, irq, resetIn, cpu_pause, pRead;
input [31:0] cpu_ib;
input [7:0] cpu_db;
output cpu_ack, vic_ack, cpu_rw;
output [15:0] cpu_a;
output [15:2] cpu_al;
output [7:0] cpu_wb;
output [3:0] cpu_wbyte;

input dma_req, dma_rw;
input [15:0] dma_a;
input [7:0] dma_dw;
output dma_ack;

`ifdef debug
input [15:0] break_pc;
input [4:0] debug_add;
input [2:0] debug_sel;
output [47:0] debug_Dout;
output breakTrig;
`endif

`ifdef pport
//input EppDstb;
input [5:0] pp_add;
output [7:0] pp_Dout;
`endif

`ifndef N6502
//****************************************************************************
output cpu_hiram, cpu_loram, cpu_charen;

//dma cannot write this register, it can write to the ram image though
assign cpu_w = ~(|cpu_a[15:1]) & cpu_a[0] & ~cpu_rw & ~cpu_ack;

//C64 cpu memory control register
reg [2:0] a1 = 3'h6;
assign cpu_loram = a1[0];
assign cpu_hiram = a1[1];
assign cpu_charen = a1[2];

always@(posedge clk)
  begin
    //update the cpu memory map control register
    if(~resetIn)   a1 <= 3'h6; //set it wrong just to know it changes
    else if(cpu_w) a1 <= cpu_wb[2:0];
  end
`endif             

//****************************************************************************
// parameter definitions
parameter V_NMI = 16'hFFFA;
parameter V_RST = 16'hFFFC;
parameter V_IRQ = 16'hFFFE;
parameter ZERO = 8'h00;
parameter ONE = 8'h01;
parameter NEG_ONE = 8'hFF;
parameter H = 1'b1;
parameter L = 1'b0;
parameter RD = 1'b1;
parameter WR = 1'b0;
parameter TRUE = 1'b1;
parameter FALSE = 1'b0;
parameter BYTEX = 4'hF;
parameter BYTE0 = 4'hE;
parameter BYTE1 = 4'hD;
parameter BYTE2 = 4'hB;
parameter BYTE3 = 4'h7;

//****************************************************************************
//cpu state machine state names

// put reset vector on address bus
parameter S_RESET = 5'h00;
// wait for new execution path (PC)
parameter S_PC_DELAY = 5'h01;
// load pc
parameter S_LOAD_PC = 5'h02;
// wait for first instruction of new code path
parameter S_INST_DELAY = 5'h03;
// load IR with first instruction bytes (start decoding instruction)
parameter S_LOAD_INST = 5'h04;
// execute 1 byte instruction
parameter S_EXECUTE_INST = 5'h05;
// execute 2 byte instruction
parameter S_EXECUTE_INST2 = 5'h06;
// execute 3 byte instruction
parameter S_EXECUTE_INST3 = 5'h07;
// memory operand load delay
parameter S_MEM_DELAY = 5'h08;
// memory write back operation
parameter S_MEM_OP = 5'h09;
// register write operation
parameter S_REG_OP = 5'h0A;
// indirect x operand address load
parameter S_IND_X = 5'h0B;
// indirect x write operation
parameter S_IND_X2 = 5'h0C;
// indirect y operand address load
parameter S_IND_Y = 5'h0D;
// indirect y operand address adjust (+y)
parameter S_IND_Y2 = 5'h0E;
// rti pc load
parameter S_RTI = 5'h0F;
// rts pc load
parameter S_RTS = 5'h10;
// rts pc adjust (+1)
parameter S_RTS2 = 5'h11;
// ZP memory write back operation
parameter S_ZP_MEM_OP = 5'h12;
// ZP register write operation
parameter S_ZP_REG_OP = 5'h13;
// give BCD arithmatic an extra cycle to complete
parameter S_BCD_DELAY = 5'h14;
// process nmi
parameter S_NMI = 5'h15;
// process irq
parameter S_IRQ = 5'h16;
// reload instruction register - ir may be corrupted by self modifying code
parameter S_IR_W_HIT = 5'h17;
// write an absolute operand
parameter S_WRITE = 5'h18;
// used to slow down absolute mem ops that hit the zp/stack range
parameter S_ZP_MEM_DELAY = 5'h19;


//state machine
reg [4:0] next_state, state = S_RESET;

//****************************************************************************
//State machine patterns...
//
// all sequences end going to one of the following states...
// S_EXECUTE_INST, S_EXECUTE_INST2, S_EXECUTE_INST3, S_IRQ, S_NMI or S_IR_W_HIT *** 
//
// S_IR_W_HIT will only be entered if an instruction's terminating state is a write
// to a location that may be in or already on its way to the ir queue. Interrupts
// override and negate this requirement.
//
// inactive default cpu_a  address {cpu_if, L} XXX
// inactive default cpu_aw address {ONE, spp1} XXX
//
// Active {cpu_if, L} accesses will only use the data and increment itself if there
// would be room in the ir register by the time the current instruction completes.
//
// # indicates a write cycle to the stack or zero page.
// % indicates a write cycle to external memoyy.
//
//****************************************************************************
// EXCEPTIONAL EVENTS :

// reset :
//
// state  : S_RESET -> S_PC_DELAY -> S_LOAD_PC -> S_INST_DELAY -> S_LOAD_INST ->> ***
// cpu_a  : <V_RST>    XXX           <pc>         {cpu_if, L}     {cpu_if, L}
// cpu_aw : XXX        XXX           XXX          XXX             XXX

// irq :
//
// state  : S_IRQ -> S_PC_DELAY -> S_LOAD_PC -> S_INST_DELAY -> S_LOAD_INST ->> ***
// cpu_a  : <V_IRQ>  XXX           <pc>         {cpu_if, L}     {cpu_if, L}
// cpu_aw : XXX      XXX           {ONE, spp3}# {ONE, spp2}#    {ONE, spp1}#

// nmi :
//
// state  : S_NMI -> S_PC_DELAY -> S_LOAD_PC -> S_INST_DELAY -> S_LOAD_INST ->> ***
// cpu_a  : <V_NMI>  XXX           <pc>         {cpu_if, L}     {cpu_if, L}
// cpu_aw : XXX      XXX           {ONE, spp3}# {ONE, spp2}#    {ONE, spp1}#

// ir refill due to self modifying code :
//
// state  : S_IR_W_HIT -> S_INST_DELAY -> S_LOAD_INST ->> ***
// cpu_a  : <pc>          {cpu_if, L}     {cpu_if, L}
// cpu_aw : XXX           XXX             XXX

//****************************************************************************
// ONE BYTE INSTRUCTIONS :

// brk :
//
// state  : S_EXECUTE_INST -> S_PC_DELAY -> S_LOAD_PC -> S_INST_DELAY -> S_LOAD_INST ->> ***
// cpu_a  : <V_IRQ>           XXX           <pc>         {cpu_if, L}     {cpu_if, L}
// cpu_aw : XXX               XXX           {ONE, spp3}# {ONE, spp2}#    {ONE, spp1}#

// rti :
//
// state  : S_EXECUTE_INST -> S_RTI -> S_INST_DELAY -> S_LOAD_INST ->> ***
// cpu_a  : XXX               <pc>     {cpu_if, L}     {cpu_if, L}
// cpu_aw : {ONE, spp1}       XXX      XXX             XXX

// rts :  in S_RTS the pc off the stack is incremented
//
// state  : S_EXECUTE_INST -> S_RTS -> S_RTS2 -> S_INST_DELAY -> S_LOAD_INST ->> ***
// cpu_a  : XXX               XXX      <pc>      {cpu_if, L}     {cpu_if, L}
// cpu_aw : {ONE, spp1}       XXX      XXX       XXX             XXX

// php, pha :
//
// state  : S_EXECUTE_INST ->> ***
// cpu_a  : {cpu_if, L}
// cpu_aw : {ONE, sp}#

// pla : in S_ZP_REG_OP flags are updated
//
// state  : S_EXECUTE_INST -> S_ZP_REG_OP ->> ***
// cpu_a  : {cpu_if, L}       {cpu_if, L}
// cpu_aw : {ONE, spp1}       XXX

// plp :
//
// state  : S_EXECUTE_INST ->> ***
// cpu_a  : {cpu_if, L}
// cpu_aw : {ONE, spp1}

// all other one byte instructions :
//
// state  : S_EXECUTE_INST ->> ***
// cpu_a  : {cpu_if, L}
// cpu_aw : XXX


//****************************************************************************
// TWO BYTE INSTRUCTIONS :
//
// branch_taken :
//
// state  : S_EXECUTE_INST2 -> S_INST_DELAY -> S_LOAD_INST ->> ***
// cpu_a  : <branch>           {cpu_if, L}     {cpu_if, L}
// cpu_aw : XXX                XXX             XXX

// (_ix) instructions : case 1 external read
//
// state  : S_EXECUTE_INST -> S_IND_X -> S_MEM_DELAY -> S_REG_OP ->> ***
// cpu_a  : {cpu_if, L}       <tmp_ixy>  {cpu_if, L}    {cpu_if, L} 
// cpu_aw : {ZERO,addx[7:0]}  <tmp_ixy>  XXX            XXX

// (_ix) instructions : case 2 zp read
//
// state  : S_EXECUTE_INST -> S_IND_X -> S_ZP_REG_OP ->> ***
// cpu_a  : {cpu_if, L}       <tmp_ixy>  {cpu_if, L}
// cpu_aw : {ZERO,addx[7:0]}  <tmp_ixy>  XXX

// (_ix) instructions : case 3 write
//
// state  : S_EXECUTE_INST -> S_IND_X ->> ***
// cpu_a  : {cpu_if, L}       <tmp_ixy>%
// cpu_aw : {ZERO,addx[7:0]}  <tmp_ixy>#

// (_iy) instructions : case 1 external read - S_IND_Y used to add the y index
//
// state  : S_EXECUTE_INST -> S_IND_Y -> S_IND_Y2 -> S_MEM_DELAY -> S_REG_OP ->> ***
// cpu_a  : {cpu_if, L}       {cpu_if, L}<tmp_ixy>   {cpu_if, L}    {cpu_if, L} 
// cpu_aw : {ZERO,add[7:0]}   XXX        <tmp_ixy>   XXX            XXX

// (_iy) instructions : case 2 zp read - S_IND_Y used to add the y index
//
// state  : S_EXECUTE_INST -> S_IND_Y -> S_IND_Y2 -> S_ZP_REG_OP ->> ***
// cpu_a  : {cpu_if, L}       {cpu_if, L}<tmp_ixy>   {cpu_if, L}
// cpu_aw : {ZERO,add[7:0]}   XXX        <tmp_ixy>    XXX

// (_iy) instructions : case 3 write - S_IND_Y used to add the y index
//
// state  : S_EXECUTE_INST -> S_IND_Y -> S_IND_Y2 ->> ***
// cpu_a  : {cpu_if, L}       {cpu_if, L}<tmp_ixy>%
// cpu_aw : {ZERO,add[7:0]}   XXX        <tmp_ixy>#

// _z : memory operations
//
// state  : S_EXECUTE_INST -> S_ZP_MEM_OP ->> ***
// cpu_a  : {cpu_if, L}       {ZERO,add[7:0]}%
// cpu_aw : {ZERO,add[7:0]}   {ZERO,add[7:0]}#

// _z : register operations
//
// state  : S_EXECUTE_INST -> S_ZP_REG_OP ->> ***
// cpu_a  : {cpu_if, L}       {cpu_if, L}
// cpu_aw : {ZERO,add[7:0]}   XXX

// _z : register operations in decimal mode
//
// state  : S_EXECUTE_INST -> S_ZP_REG_OP  -> S_BCD_DELAY ->> ***
// cpu_a  : {cpu_if, L}       {cpu_if, L}     {cpu_if, L}
// cpu_aw : {ZERO,add[7:0]}   XXX             XXX

// _zx : memory operations
//
// state  : S_EXECUTE_INST -> S_ZP_MEM_OP ->> ***
// cpu_a  : {cpu_if, L}       {ZERO,addx[7:0]}%
// cpu_aw : {ZERO,addx[7:0]}  {ZERO,addx[7:0]}#

// _zx : register operations
//
// state  : S_EXECUTE_INST -> S_ZP_REG_OP ->> ***
// cpu_a  : {cpu_if, L}       {cpu_if, L}
// cpu_aw : {ZERO,addx[7:0]}  XXX

// _zx : register operations in decimal mode
//
// state  : S_EXECUTE_INST -> S_ZP_REG_OP  -> S_BCD_DELAY ->> ***
// cpu_a  : {cpu_if, L}       {cpu_if, L}     {cpu_if, L}
// cpu_aw : {ZERO,addx[7:0]}  XXX             XXX

`ifdef zps_w_thru

// sta_z,stx_z,sty_z :
//
// state  : S_EXECUTE_INST -> S_WRITE ->> ***
// cpu_a  : {cpu_if, L}       {ZERO,add[7:0]}%
// cpu_aw : XXX               {ZERO,add[7:0]}#

// sta_zx,sty_zx :
//
// state  : S_EXECUTE_INST -> S_WRITE ->> ***
// cpu_a  : {cpu_if, L}       {ZERO,addx[7:0]}%
// cpu_aw : XXX               {ZERO,addx[7:0]}#

// stx_zy :
//
// state  : S_EXECUTE_INST -> S_WRITE ->> ***
// cpu_a  : {cpu_if, L}       {ZERO,addy[7:0]}%
// cpu_aw : XXX               {ZERO,addy[7:0]}#

`else

// sta_z,stx_z,sty_z :
//
// state  : S_EXECUTE_INST ->> ***
// cpu_a  : {cpu_if, L}
// cpu_aw : {ZERO,add[7:0]}#

// sta_zx,sty_zx :
//
// state  : S_EXECUTE_INST ->> ***
// cpu_a  : {cpu_if, L}
// cpu_aw : {ZERO,addx[7:0]}#

// stx_zy :
//
// state  : S_EXECUTE_INST ->> ***
// cpu_a  : {cpu_if, L}
// cpu_aw : {ZERO,addy[7:0]}#

`endif

// ldx_zy :
//
// state  : S_EXECUTE_INST -> S_ZP_REG_OP ->> ***
// cpu_a  : {cpu_if, L}       {cpu_if, L}
// cpu_aw : {ZERO,addy[7:0]}  XXX

// all immediate decimal mode instructions :
//
// state  : S_EXECUTE_INST -> S_BCD_DELAY ->> ***
// cpu_a  : {cpu_if, L}       {cpu_if, L}
// cpu_aw : XXX               XXX

// all other immediate instructions :
//
// state  : S_EXECUTE_INST ->> ***
// cpu_a  : {cpu_if, L}
// cpu_aw : XXX

//****************************************************************************
// THREE BYTE INSTRUCTIONS :

// jmp_i :
//
// state  : S_EXECUTE_INST3 -> S_PC_DELAY -> S_LOAD_PC -> S_INST_DELAY -> S_LOAD_INST ->> ***
// cpu_a  : <add>              XXX           <pc>         {cpu_if, L}     {cpu_if, L}
// cpu_aw : XXX                XXX           XXX          XXX             XXX

// jmp :
//
// state  : S_EXECUTE_INST2 -> S_INST_DELAY -> S_LOAD_INST ->> ***
// cpu_a  : <add>              {cpu_if, L}     {cpu_if, L}
// cpu_aw : XXX                XXX             XXX

// jsr :
//
// state  : S_EXECUTE_INST2 -> S_INST_DELAY -> S_LOAD_INST ->> ***
// cpu_a  : <add>              {cpu_if, L}     {cpu_if, L}
// cpu_aw : XXX                {ONE, spp2}#    {ONE, spp1}#

// _a : memory operations - external
//
// state  : S_EXECUTE_INST -> S_MEM_DELAY -> S_MEM_OP  ->> ***
// cpu_a  : {add}             {cpu_if, L}    <tmp_a>%
// cpu_aw : {add}             XXX            <tmp_a>#

// _a : register operations - external
//
// state  : S_EXECUTE_INST -> S_MEM_DELAY -> S_REG_OP ->> ***
// cpu_a  : {add}             {cpu_if, L}    {cpu_if, L}
// cpu_aw : {add}             XXX            XXX

// _a : register operations in decimal mode - external
//
// state  : S_EXECUTE_INST -> S_MEM_DELAY -> S_REG_OP  -> S_BCD_DELAY ->> ***
// cpu_a  : {add}             {cpu_if, L}    {cpu_if, L}  {cpu_if, L}
// cpu_aw : {add}             XXX            XXX          XXX

`ifndef zps_w_thru
// _a : memory operations - external
//
// state  : S_EXECUTE_INST -> S_ZP_MEM_DELAY -> S_MEM_OP  ->> ***
// cpu_a  : {add}             {cpu_if, L}       <tmp_a>%
// cpu_aw : {add}             XXX               <tmp_a>#
`else
// _a : memory operations - internal
//
// state  : S_EXECUTE_INST -> S_ZP_MEM_OP ->> ***
// cpu_a  : {add}             {cpu_if, L}
// cpu_aw : {add}             <tmp_a>#
`endif

// _a : register operations - internal
//
// state  : S_EXECUTE_INST -> S_ZP_REG_OP ->> ***
// cpu_a  : {add}             {cpu_if, L}
// cpu_aw : {add}              XXX

// _a : register operations in decimal mode - internal
//
// state  : S_EXECUTE_INST -> S_ZP_REG_OP  -> S_BCD_DELAY ->> ***
// cpu_a  : {add}             {cpu_if, L}     {cpu_if, L}
// cpu_aw : {add}             XXX             XXX

// _ax : memory operations - external
//
// state  : S_EXECUTE_INST -> S_MEM_DELAY -> S_MEM_OP  ->> ***
// cpu_a  : {addx}            {cpu_if, L}    <tmp_a>%
// cpu_aw : {addx}            XXX            <tmp_a>#

// _ax : register operations - external
//
// state  : S_EXECUTE_INST -> S_MEM_DELAY -> S_REG_OP ->> ***
// cpu_a  : {addx}            {cpu_if, L}    {cpu_if, L}
// cpu_aw : {addx}            XXX            XXX

// _ax : register operations in decimal mode - external
//
// state  : S_EXECUTE_INST -> S_MEM_DELAY -> S_REG_OP  -> S_BCD_DELAY ->> ***
// cpu_a  : {addx}            {cpu_if, L}    {cpu_if, L}  {cpu_if, L}
// cpu_aw : {addx}            XXX            XXX          XXX

`ifndef zps_w_thru
// _ax : memory operations - external
//
// state  : S_EXECUTE_INST -> S_ZP_MEM_DELAY -> S_MEM_OP  ->> ***
// cpu_a  : {addx}            {cpu_if, L}       <tmp_a>%
// cpu_aw : {addx}            XXX               <tmp_a>#
`else
// _ax : memory operations - internal
//
// state  : S_EXECUTE_INST -> S_ZP_MEM_OP ->> ***
// cpu_a  : {addx}            {cpu_if, L}
// cpu_aw : {addx}            <tmp_a>#
`endif

// _ax : register operations - internal
//
// state  : S_EXECUTE_INST -> S_ZP_REG_OP ->> ***
// cpu_a  : {addx}            {cpu_if, L}
// cpu_aw : {addx}             XXX

// _ax : register operations in decimal mode - internal
//
// state  : S_EXECUTE_INST -> S_ZP_REG_OP  -> S_BCD_DELAY ->> ***
// cpu_a  : {addx}            {cpu_if, L}     {cpu_if, L}
// cpu_aw : {addx}            XXX             XXX


// _ay : register operations - external
//
// state  : S_EXECUTE_INST -> S_MEM_DELAY -> S_REG_OP ->> ***
// cpu_a  : {addy}            {cpu_if, L}    {cpu_if, L}
// cpu_aw : {addy}            XXX            XXX

// _ay : register operations in decimal mode - external
//
// state  : S_EXECUTE_INST -> S_MEM_DELAY -> S_REG_OP  -> S_BCD_DELAY ->> ***
// cpu_a  : {addy}            {cpu_if, L}    {cpu_if, L}  {cpu_if, L}
// cpu_aw : {addy}            XXX            XXX          XXX

// _ay : register operations - internal
//
// state  : S_EXECUTE_INST -> S_ZP_REG_OP ->> ***
// cpu_a  : {addy}            {cpu_if, L}
// cpu_aw : {addy}             XXX

// _ay : register operations in decimal mode - internal
//
// state  : S_EXECUTE_INST -> S_ZP_REG_OP  -> S_BCD_DELAY ->> ***
// cpu_a  : {addy}           {cpu_if, L}     {cpu_if, L}
// cpu_aw : {addy}            XXX             XXX

// sta_a,stx_a,sty_a :
//
// state  : S_EXECUTE_INST -> S_WRITE ->> ***
// cpu_a  : {cpu_if, L}       {add}%
// cpu_aw : XXX               {add}#

// sta_ax :
//
// state  : S_EXECUTE_INST -> S_WRITE ->> ***
// cpu_a  : {cpu_if, L}       {addx}%
// cpu_aw : XXX               {addx}#

// sta_ay :
//
// state  : S_EXECUTE_INST -> S_WRITE ->> ***
// cpu_a  : {cpu_if, L}       {addy}%
// cpu_aw : XXX               {addy}#


//****************************************************************************
reg [7:0] ira, imm_op, ah;
//wire [7:0] irr;

iDecode id(clk, ira,
           two_byte, one_byte_a, three_byte_a,
           asla, rola, lsra, rora, dex, dey, inx, iny,txa, tya, txs, tay,
           tax, tsx, pha, pla,clc, sec, cli, sei, cld, sed, clv, php, plp,
           ora_i, anda_i, eora_i, lda_i, cmpa_i, adca_i, sbca_i,
           ora_z, anda_z, eora_z, lda_z, cmpa_z, adca_z, sbca_z, bita_z,
           ora_zx, anda_zx, eora_zx, lda_zx, cmpa_zx, adca_zx, sbca_zx,
           sta_ix, ora_ix, anda_ix, eora_ix, lda_ix, cmpa_ix, adca_ix, sbca_ix,
           sta_iy, ora_iy, anda_iy, eora_iy, lda_iy, cmpa_iy, adca_iy, sbca_iy,
           ora_a, anda_a, eora_a, lda_a, cmpa_a, bita_a, adca_a, sbca_a,
           ora_ax, anda_ax, eora_ax, ldy_ax, lda_ax, cmpa_ax, adca_ax, sbca_ax,
           ora_ay, anda_ay, eora_ay, lda_ay, ldx_ay, cmpa_ay, adca_ay, sbca_ay,
           asl_z, rol_z, lsr_z, ror_z, dec_z, inc_z,
           asl_zx, rol_zx, lsr_zx, ror_zx, dec_zx, inc_zx,
           asl_a, rol_a, lsr_a, ror_a, dec_a, inc_a,
           asl_ax, rol_ax, lsr_ax, ror_ax, dec_ax, inc_ax,
           stx_z, ldx_z, cpx_z, stx_zy, ldx_zy, stx_a, ldx_a, cpx_a, ldx_i, cpx_i,
           sty_z, ldy_z, cpy_z, sty_zx, ldy_zx, sty_a, ldy_a, cpy_a, ldy_i, cpy_i,
           bpl, bmi, bvc, bvs, bcc, bcs, bne, beq,
           brk, rti, rts, jmp_a, jmp_i, jsr_a,
           ix, iy, st_zxy, st_axy, zx, zy, ax, ay, mem_op, reg_op);
           //irr, nop, sta_z, sta_zx, sta_a, sta_ax, sta_ay);

//decoded instructions saved for use out of the S_EXECUTE_INST states
reg sta_ix_2, sta_iy_2, addsub, mem_op_2, sty, stx;
reg asl, rol, ror, lsr, inc, dec;
reg ora, bita, anda, eora, ldx, lda, ldy, cpx_2, cmpa_2, cpy_2;

//data and index registers
reg [7:0] a, x, y, alu_op;

//status register bits
reg negative, zero, carry, overflow, interrupt, decimal;

reg [15:0] next_pc;
reg [15:1] next_if;
reg [7:0] next_sp;
reg [15:0] addy;
reg [7:0] next_y;
reg [15:0] addx;
reg [7:0] next_x;
wire [7:0] next_spp1, next_spp2, next_spp3;
assign next_spp1 = next_sp + H;
assign next_spp2 = next_sp + 2'h2;
assign next_spp3 = next_sp + 2'h3;
wire [15:0] tmp_ax, tmp_ah;
assign tmp_ah = { ah, imm_op };
assign tmp_ax = tmp_ah + next_x;

//instruction registers
//bit 0 never used...accesses always on word boundaries (for 4 usable bytes)
reg [15:1] cpu_if;
wire [15:1] cpu_ifp4 = cpu_if + 15'h0002;

reg [95:0] ir;        //12 bytes of instruction queue
reg [3:0] iCount;     //how many instruction bytes currently in the ir register
reg ifCycle, ifReady; //Instruction fetch memory cycle started / completed

//ifCycle (Ready) data buses
reg [7:0] ib0,ib1,ib2,ib3;
wire [31:0] ib32;
assign ib32 = { ib3, ib2, ib1, ib0 };

//for making sure we can decode the instruction we think that we have
assign four_valid = |iCount[3:2];
assign two_valid = four_valid | iCount[1];
//assign one_valid = two_valid | iCount[0];
assign three_valid = four_valid | (&iCount[1:0]);
assign five_valid = iCount[3] | (iCount[2] & (|iCount[1:0]));
assign six_valid = iCount[3] | (&iCount[2:1]);
//assign seven_valid = iCount[3] | (&iCount[2:0]);

//if low an add is occurring else it is a sub or cmp
reg mode;

//alu outputs
wire [7:0] result, asl_op, rol_op, lsr_op, ror_op, dec_op, inc_op, or_op, and_op, eor_op, bcdResult;

//binary alu
accumulatorPlus aA(mode, a, alu_op, carry, result, Cout_a, is_zero_a, is_negative_a, is_overflow_a,
                   alu_op_0, alu_op_1, alu_op_255, alu_op_dec_neg, alu_op_inc_neg,
                   asl_op, rol_op, lsr_op, ror_op, dec_op, inc_op,
                   or_op, and_op, eor_op, or_op_zero, and_op_zero, eor_op_zero);
                   
//decimal mode alu
bcdAccumulator aBCD(mode, a, alu_op, carry, bcdResult, bcdCout_a, bcdZero_a, bcdNegative_a, bcdOverflow_a);

//compare alu
compare cA(a, alu_op, Cout_ca, is_zero_ca, N_ca);
compare cX(x, alu_op, Cout_x, is_zero_x, N_x);
compare cY(y, alu_op, Cout_y, is_zero_y, N_y);

//add branch offset to the pc instead the instruction size
assign branch_taken = bpl&~negative|bmi&negative|bvc&~overflow|bvs&overflow|bcc&~carry|bcs&carry|bne&~zero|beq&zero;

//synchronizing interrupt registers
reg rirq = H;       //synchonize irq
reg rnmi = H;       //synchonize nmi
reg rnmi2 = H;      //this is used to determine a falling rnmi edge
reg nmi = H;        //this is used to save the falling edge occurance until the next instruction

//internally held interrupt vector pc caches
//beware if your system bank switches these values - should work for C64.
//if bit 17 is zero they are disabled until initialized by an actual interrupt event
//reg [16:0] irq_rom = 17'h00000;
//reg [16:0] irq_ram = 17'h00000;
//reg [16:0] nmi_rom = 17'h00000;
//reg [16:0] nmi_ram = 17'h00000;

//an unimplemented instruction or decimal mode or some other fatal event causes at least
//1 bit in this variable to be set low and stops the processor in a halt condition
reg [3:0] halt = 4'hF;

//write bus registers
reg [7:0] cpu_wb;
reg cpu_rw, cpu_rwz;

//byte write control registers
reg [3:0] cpu_wbyte, cpu_wbytez;

//address registers
reg [15:0] pc, pc_p2, branch, tmp_a, cpu_a, cpu_aw, tmp_pc, tmp_ixy;
//reading an address from a memory
reg [15:0] cpu_dar0, cpu_dar1, cpu_dar2, cpu_dar3;
//program counter increment options
wire [15:0] pc_p1, pc_p3, pcp2, add;

//operand address
assign add = ir[23:8];

assign pc_p1 = pc + 16'h0001;
assign pc_p3 = pc + 16'h0003;
assign pcp2 = next_pc + 16'h0002;

//stack / zero page cache
assign cacherw_hit = ~(|cpu_aw[15:9]);
reg [1:0] cacheiSelect, opSelect;

wire [7:2] cpu_awl;
wire [31:0] cache, caches, cachez;
ram256 zerop(clk, cpu_wbytez, ~cpu_aw[8] & cacherw_hit, cpu_aw[7:2], cpu_awl, { cpu_wb, cpu_wb, cpu_wb, cpu_wb }, cachez);
ram256 stack(clk, cpu_wbytez, cpu_aw[8] & cacherw_hit, cpu_aw[7:2], cpu_awl, { cpu_wb, cpu_wb, cpu_wb, cpu_wb }, caches);
assign cache = cpu_aw[8] ? caches : cachez;

//stack pointers used for address bus
reg [7:0] sp, spp1, spp2, spp3;
//stack pointer variations
wire [7:0] spm1, spm2, spm3;

assign spm1 = sp - H;
assign spm2 = sp - 2'h2;
assign spm3 = sp - 2'h3;

//processor status stacking options
wire [7:0] sp_int, sp_brk;
assign sp_int = { negative, overflow, H, L, decimal, interrupt, zero, carry };
assign sp_brk = { negative, overflow, H, H, decimal, interrupt, zero, carry };

//read registers, saves data from the data bus during a vic/dma cycle and holds
//it for processing during the next available cpu cycle...
reg [7:0] cpu_dbr;
reg [31:0] cpu_ir, cpu_ibr, cachei_r;

//acknowledge signals for bus mastership
reg cpu_ack2, vic_ack, dma_ack, cpu_ack;

reg [24:0] stackit;

`ifdef debug
reg [31:0] cpu_drr;
reg [15:0] cpu_a2_i, cpu_aw2_i;
reg [7:0] cpu_pr, cpu_prr, cpu_wb_r;
reg [3:0] cpu_byte_enable;
reg pReadr, cpu_rw_r;
//for pausing cpu due to certain events - starts single step
reg breakTrig = FALSE;
reg breakTrig1 = FALSE;
//for counting events
reg [31:0] cycleCount;
`endif

//asynchronous control state machines
//****************************************************************************
//****************************************************************************
//ir decode
always@(*)
  begin
    if(~cpu_ack)
      begin
        case(state)
//****************************************************************************
          S_LOAD_INST:
            begin
              case(pc[1:0])
                2'h0 :
                  begin
                    ira = cpu_ir[7:0];
                    imm_op = cpu_ir[15:8];
                    ah = cpu_ir[23:16];
                  end
                2'h1 :
                  begin
                    ira = cpu_ir[15:8];
                    imm_op = cpu_ir[23:16];
                    ah = cpu_ir[31:24];
                  end
                2'h2 :
                  begin
                    ira = cpu_ir[23:16];
                    imm_op = cpu_ir[31:24];
                    ah = cpu_ir[7:0];
                  end
                2'h3 :
                  begin
                    ira = cpu_ir[31:24];
                    imm_op = cpu_ir[7:0];
                    ah = cpu_ir[15:8];
                  end
              endcase
            end
//****************************************************************************
          S_EXECUTE_INST:
            begin
              if(two_valid) ira = ir[15:8];
              else          ira = ib0;

              if(four_valid)
               begin
                ah = ir[31:24];
                imm_op = ir[23:16];
               end
             else if(three_valid)
               begin
                 ah = ib0;
                 imm_op = ir[23:16];
               end
             else if(two_valid)
               begin
                 ah = ib1;
                 imm_op = ib0;
               end
             else
               begin
                 ah = ib2;
                 imm_op = ib1;
               end
            end
//****************************************************************************
          S_EXECUTE_INST2:
            begin
              if(three_valid) ira = ir[23:16];
              else            ira = ib0;

              if(five_valid)
                begin
                  ah = ir[39:32];
                  imm_op = ir[31:24];
                end
              else if(four_valid)
                begin
                  ah = ib0;
                  imm_op = ir[31:24];
                end
              else if(three_valid)
                begin
                  ah = ib1;
                  imm_op = ib0;
                end
              else
                begin
                 ah = ib2;
                 imm_op = ib1;
                end
            end
//****************************************************************************
          S_EXECUTE_INST3:
            begin
              if(four_valid) ira = ir[31:24];
              else           ira = ib0;

              if(six_valid)
                begin
                  ah = ir[47:40];
                  imm_op = ir[39:32];
                end
              else if(five_valid)
                begin
                  ah = ib0;
                  imm_op = ir[39:32];
                end
              else if(four_valid)
                begin
                  ah = ib1;
                  imm_op = ib0;
                end
              else
                begin
                  ah = ib2;
                  imm_op = ib1;
                end
            end
//****************************************************************************
          default:
            begin
              ira = ir[7:0];
              imm_op = ir[15:8];
              ah = ir[23:16];
            end
        endcase
      end   
    else
      begin
        ira = ir[7:0];
        imm_op = ir[15:8];
        ah = ir[23:16];
      end
  end

//****************************************************************************
//****************************************************************************
//write bus
always@(*)
  begin
    if(~cpu_ack)
      begin
        case(state)
//****************************************************************************
          S_LOAD_PC:                cpu_wb = stackit[23:16];
//****************************************************************************
          S_INST_DELAY:             cpu_wb = stackit[15:8];
//****************************************************************************
          S_LOAD_INST:              cpu_wb = stackit[7:0];
//****************************************************************************
          S_EXECUTE_INST:
            begin
              if(php)               cpu_wb = sp_brk;
              else                  cpu_wb = a;
            end
`ifndef zps_w_thru
//****************************************************************************
          S_EXECUTE_INST2:
            begin
              if(sty_z|sty_zx)      cpu_wb = y;
              else if(stx_z|stx_zy) cpu_wb = x;
              else                  cpu_wb = a;
            end
`endif
//****************************************************************************
          S_MEM_OP,S_ZP_MEM_OP: // complete memory operation
            begin
              if(asl)               cpu_wb = asl_op;
              else if(rol)          cpu_wb = rol_op;
              else if(lsr)          cpu_wb = lsr_op;
              else if(ror)          cpu_wb = ror_op;
              else if(dec)          cpu_wb = dec_op;
              else                  cpu_wb = inc_op;
            end
//****************************************************************************
          S_WRITE:
            begin
              if(sty)               cpu_wb = y;
              else if(stx)          cpu_wb = x;
              else                  cpu_wb = a;
            end
//****************************************************************************
          default:                  cpu_wb = a;
        endcase
      end
    else                            cpu_wb = dma_dw;
  end


//****************************************************************************
//****************************************************************************
//r/w control (of the write bus)
always@(*)
  begin
    if(~cpu_ack)
      begin
        case(state)
`ifndef zps_w_thru
//****************************************************************************
          S_EXECUTE_INST2:                    cpu_rw = ~st_zxy;
`endif
//****************************************************************************
          S_IND_X:                            cpu_rw = ~sta_ix_2;
//****************************************************************************
          S_IND_Y2:                           cpu_rw = ~sta_iy_2;
//****************************************************************************
          S_MEM_OP,S_ZP_MEM_OP,S_WRITE:       cpu_rw = WR;
//****************************************************************************
          default:                            cpu_rw = RD;
        endcase
      end
    else if(~dma_ack)                         cpu_rw = dma_rw;
    else                                      cpu_rw = RD;
  end


//****************************************************************************
//****************************************************************************
//write byte select
always@(*)
  begin
    if(~cpu_rw)
      begin
        case(cpu_a[1:0])
          2'h0 : cpu_wbyte = BYTE0;
          2'h1 : cpu_wbyte = BYTE1;
          2'h2 : cpu_wbyte = BYTE2;
          2'h3 : cpu_wbyte = BYTE3;
        endcase
      end
    else         cpu_wbyte = BYTEX;
  end

//****************************************************************************
//****************************************************************************
//read instruction, pc vector address bus
always@(*)
  begin
    if(~cpu_ack)
      begin
        case(state)
//****************************************************************************
          S_RESET:                      cpu_a = V_RST;
//****************************************************************************
          S_LOAD_PC:
            begin
              case(tmp_a[1:0])
                2'h0 :                  cpu_a = cpu_dar0;
                2'h1 :                  cpu_a = cpu_dar1;
                2'h2 :                  cpu_a = cpu_dar2;
                2'h3 :                  cpu_a = cpu_dar3;
              endcase
            end
//****************************************************************************
          S_EXECUTE_INST:
            begin
              if(brk)                   cpu_a = V_IRQ;
              else                      cpu_a = { cpu_if, L };
            end
//****************************************************************************
          S_EXECUTE_INST2:
            begin
              if(branch_taken)          cpu_a = branch;
              else                      cpu_a = { cpu_if, L };
            end
//****************************************************************************
          S_EXECUTE_INST3:
            begin
              if(st_axy)                cpu_a = { cpu_if, L };
              else if(ax)               cpu_a = addx;
              else if(ay)               cpu_a = addy;
              else                      cpu_a = add;
            end
//****************************************************************************
          S_IND_X,S_IND_Y2:             cpu_a = tmp_ixy;
//****************************************************************************
          S_RTI:                        cpu_a = tmp_pc;
//****************************************************************************
          S_RTS:                        cpu_a = { ONE, spp1 };
//****************************************************************************
          S_RTS2:                       cpu_a = tmp_pc;
//****************************************************************************
          S_NMI:                        cpu_a = V_NMI;
//****************************************************************************
          S_IRQ:                        cpu_a = V_IRQ;
//****************************************************************************
          S_IR_W_HIT:                   cpu_a = pc;
//****************************************************************************
          S_MEM_OP,S_ZP_MEM_OP,S_WRITE: cpu_a = tmp_a;
//****************************************************************************
          default:                      cpu_a = { cpu_if, L };
        endcase
      end
    else if(~dma_ack)                   cpu_a = dma_a;
    else                                cpu_a = { vic_a, L };
  end

//lower 16 bit data bus address generation 
assign cpu_al = cpu_a[15:2] + cpu_a[1];

//****************************************************************************
//****************************************************************************
//r/w control (of the write bus)
always@(*)
  begin
    if(~cpu_ack)
      begin
        case(state)
//****************************************************************************
          S_LOAD_PC,S_INST_DELAY,S_LOAD_INST: cpu_rwz = ~stackit[24];
//****************************************************************************
          S_EXECUTE_INST:                     cpu_rwz = ~(pha|php);
`ifndef zps_w_thru
//****************************************************************************
          S_EXECUTE_INST2:                    cpu_rwz = ~st_zxy;
`endif
//****************************************************************************
          S_IND_X:                            cpu_rwz = ~sta_ix_2;
//****************************************************************************
          S_IND_Y2:                           cpu_rwz = ~sta_iy_2;
//****************************************************************************
          S_MEM_OP,S_ZP_MEM_OP,S_WRITE:       cpu_rwz = WR;
//****************************************************************************
          default:                            cpu_rwz = RD;
        endcase
      end
    else if(~dma_ack)                         cpu_rwz = dma_rw;
    else                                      cpu_rwz = RD;
  end


//****************************************************************************
//****************************************************************************
//write byte select
always@(*)
  begin
    if(~cpu_rwz)
      begin
        case(cpu_aw[1:0])
          2'h0 : cpu_wbytez = BYTE0;
          2'h1 : cpu_wbytez = BYTE1;
          2'h2 : cpu_wbytez = BYTE2;
          2'h3 : cpu_wbytez = BYTE3;
        endcase
      end
    else         cpu_wbytez = BYTEX;
  end

//****************************************************************************
//****************************************************************************
// write or read instruction operand address bus
always@(*)
  begin
    if(~cpu_ack)
      begin
        case(state)
//****************************************************************************
          S_LOAD_PC:                    cpu_aw = { ONE, spp3 };
//****************************************************************************
          S_INST_DELAY:                 cpu_aw = { ONE, spp2 };
//****************************************************************************
          S_EXECUTE_INST:
            begin
              if(pha|php)               cpu_aw = { ONE, sp };
              else                      cpu_aw = { ONE, spp1 };
            end
//****************************************************************************
          S_EXECUTE_INST2:
            begin
              if(zx|ix)                 cpu_aw = { ZERO, addx[7:0] };
              else if(zy)               cpu_aw = { ZERO, addy[7:0] };
              else                      cpu_aw = { ZERO, add[7:0] };
            end
//****************************************************************************
          S_EXECUTE_INST3:
            begin
              if(st_axy)                cpu_aw = { ONE, spp1 };
              else if(ax)               cpu_aw = addx;
              else if(ay)               cpu_aw = addy;
              else                      cpu_aw = add;
            end
//****************************************************************************
          S_IND_X,S_IND_Y2:             cpu_aw = tmp_ixy;
//****************************************************************************
          S_MEM_OP,S_ZP_MEM_OP,S_WRITE: cpu_aw = tmp_a;
//****************************************************************************
          default:                      cpu_aw = { ONE, spp1 };
        endcase
      end
    else                                cpu_aw = dma_a;
  end

//lower 16 bit cache data bus address generation 
assign cpu_awl = cpu_aw[7:2] + cpu_aw[1];

//****************************************************************************
//****************************************************************************
//next y look-ahead
always@(*)
  begin
    case(state)
//****************************************************************************
      S_EXECUTE_INST:
        begin
          if(tay)      next_y = a;
          else if(dey) next_y = y - ONE;
          else if(iny) next_y = y + ONE;
          else         next_y = y;
        end
//****************************************************************************
      S_EXECUTE_INST2:
        begin
          if(ldy_i)    next_y = ir[15:8];
          else         next_y = y;
        end
//****************************************************************************
      S_REG_OP, S_ZP_REG_OP:
        if(ldy)        next_y = alu_op;
        else           next_y = y;
//****************************************************************************
      default:         next_y = y;
    endcase
  end

//****************************************************************************
//****************************************************************************
//next x look-ahead
always@(*)
  begin
    case(state)
//****************************************************************************
      S_EXECUTE_INST:
        begin
          if(tax)      next_x = a;
          else if(tsx) next_x = sp;
          else if(dex) next_x = x - ONE;
          else if(inx) next_x = x + ONE;
          else         next_x = x;
        end
//****************************************************************************
      S_EXECUTE_INST2:
        begin
          if(ldx_i)    next_x = ir[15:8];
          else         next_x = x;
        end
//****************************************************************************
      S_REG_OP, S_ZP_REG_OP:
        if(ldx)        next_x = alu_op;
        else           next_x = x;
//****************************************************************************
      default:         next_x = x;
    endcase
  end

//****************************************************************************
//****************************************************************************
//next state look-ahead
always@(*)
  begin
    case(state)
//****************************************************************************
      S_RESET:                                     next_state = S_PC_DELAY;  
//****************************************************************************
      S_PC_DELAY:                                  next_state = S_LOAD_PC;  
//****************************************************************************
      S_LOAD_PC:                                   next_state = S_INST_DELAY;  
//****************************************************************************
      S_INST_DELAY:                                next_state = S_LOAD_INST;  
//****************************************************************************
      S_LOAD_INST:
        if(one_byte_a)                             next_state = S_EXECUTE_INST;
        else if(three_byte_a)                      next_state = S_EXECUTE_INST3;
        else                                       next_state = S_EXECUTE_INST2;
//****************************************************************************
      S_EXECUTE_INST:
        if(brk)                                    next_state = S_PC_DELAY;
        else if(rti)                               next_state = S_RTI;
        else if(rts)                               next_state = S_RTS;
        else if(pla)                               next_state = S_ZP_REG_OP;
        else if(~nmi)                              next_state = S_NMI;
        else if(~rirq)                             next_state = S_IRQ;
        else if(one_byte_a)                        next_state = S_EXECUTE_INST;
        else if(three_byte_a)                      next_state = S_EXECUTE_INST3;
        else                                       next_state = S_EXECUTE_INST2;
//****************************************************************************
      S_EXECUTE_INST2:
        if(branch_taken)                           next_state = S_INST_DELAY;
`ifdef zps_w_thru
        else if(st_zxy)                            next_state = S_WRITE;
`endif
        else if(ix)                                next_state = S_IND_X;
        else if(iy)                                next_state = S_IND_Y;
        else if(mem_op)                            next_state = S_ZP_MEM_OP;
        else if(reg_op)                            next_state = S_ZP_REG_OP;
        else if(addsub & decimal)                  next_state = S_BCD_DELAY;
        else if(~nmi)                              next_state = S_NMI;
        else if(~rirq)                             next_state = S_IRQ;
        else if(((cpu_aw - pc_p2) < 16) & ~cpu_rw) next_state = S_IR_W_HIT;
        else if(one_byte_a)                        next_state = S_EXECUTE_INST;
        else if(three_byte_a)                      next_state = S_EXECUTE_INST3;
        else                                       next_state = S_EXECUTE_INST2;
//****************************************************************************
      S_EXECUTE_INST3:
        if(jsr_a|jmp_a)                            next_state = S_INST_DELAY; 
        else if(jmp_i)                             next_state = S_PC_DELAY; 
        else if(st_axy)                            next_state = S_WRITE; 
        else if(cacherw_hit) //turn absolute access into a stack or zp access
          begin
`ifdef zps_w_thru
            // in this case the internal memory will be used but not until after the
            //next state like a normal external memory delay - it is possible the
            //external data is invalid because of direct stack pushes...
            if(mem_op)                             next_state = S_ZP_MEM_DELAY;
`else
            if(mem_op)                             next_state = S_ZP_MEM_OP;
`endif
            else                                   next_state = S_ZP_REG_OP;
          end
        else                                       next_state = S_MEM_DELAY;
//****************************************************************************
      S_MEM_DELAY,S_ZP_MEM_DELAY: // reading memory operand
        if(mem_op_2)                               next_state = S_MEM_OP; 
        else                                       next_state = S_REG_OP; 
//****************************************************************************
      S_IND_X:  // indirect x
        if(~sta_ix_2)
          begin
            //turn absolute access into a stack or zp access
            if(cacherw_hit)                        next_state = S_ZP_REG_OP;
            else                                   next_state = S_MEM_DELAY;
          end
        else if(~nmi)                              next_state = S_NMI;
        else if(~rirq)                             next_state = S_IRQ;
        else if(((cpu_aw - pc) < 16) & ~cpu_rw)    next_state = S_IR_W_HIT;
        else if(one_byte_a)                        next_state = S_EXECUTE_INST;
        else if(three_byte_a)                      next_state = S_EXECUTE_INST3;
        else                                       next_state = S_EXECUTE_INST2;
//****************************************************************************
      S_IND_Y:                                     next_state = S_IND_Y2;
//****************************************************************************
      S_IND_Y2:  // indirect y
        if(~sta_iy_2)
          begin
            //turn absolute access into a stack or zp access
            if(cacherw_hit)                        next_state = S_ZP_REG_OP;
            else                                   next_state = S_MEM_DELAY;
          end
        else if(~nmi)                              next_state = S_NMI;
        else if(~rirq)                             next_state = S_IRQ;
        else if(((cpu_aw - pc) < 16) & ~cpu_rw)    next_state = S_IR_W_HIT;
        else if(one_byte_a)                        next_state = S_EXECUTE_INST;
        else if(three_byte_a)                      next_state = S_EXECUTE_INST3;
        else                                       next_state = S_EXECUTE_INST2;
//****************************************************************************
      S_MEM_OP ,S_ZP_MEM_OP:
        if(~nmi)                                   next_state = S_NMI;
        else if(~rirq)                             next_state = S_IRQ;
        else if(((cpu_aw - pc) < 16) & ~cpu_rw)    next_state = S_IR_W_HIT;
        else if(one_byte_a)                        next_state = S_EXECUTE_INST;
        else if(three_byte_a)                      next_state = S_EXECUTE_INST3;
        else                                       next_state = S_EXECUTE_INST2;
//****************************************************************************
      S_REG_OP, S_ZP_REG_OP:
        if(addsub & decimal)                       next_state = S_BCD_DELAY;
        else if(~nmi)                              next_state = S_NMI;
        else if(~rirq)                             next_state = S_IRQ;
        else if(one_byte_a)                        next_state = S_EXECUTE_INST;
        else if(three_byte_a)                      next_state = S_EXECUTE_INST3;
        else                                       next_state = S_EXECUTE_INST2;
//****************************************************************************
      S_RTI:                                       next_state = S_INST_DELAY;
//****************************************************************************
      S_RTS:                                       next_state = S_RTS2;
//****************************************************************************
      S_RTS2:                                      next_state = S_INST_DELAY;
//****************************************************************************
      S_BCD_DELAY:
        if(~nmi)                                   next_state = S_NMI;
        else if(~rirq)                             next_state = S_IRQ;
        else if(one_byte_a)                        next_state = S_EXECUTE_INST;
        else if(three_byte_a)                      next_state = S_EXECUTE_INST3;
        else                                       next_state = S_EXECUTE_INST2;
//****************************************************************************
      S_NMI, S_IRQ:                                next_state = S_PC_DELAY;
//****************************************************************************
      S_IR_W_HIT:                                  next_state = S_INST_DELAY;
//****************************************************************************
      S_WRITE:
        if(~nmi)                                   next_state = S_NMI;
        else if(~rirq)                             next_state = S_IRQ;
        else if((cpu_aw - pc) < 16)                next_state = S_IR_W_HIT;
        else if(one_byte_a)                        next_state = S_EXECUTE_INST;
        else if(three_byte_a)                      next_state = S_EXECUTE_INST3;
        else                                       next_state = S_EXECUTE_INST2;
//****************************************************************************
      default:                                     next_state = state; //should never get here
    endcase
  end

//****************************************************************************
//****************************************************************************
//next pc look-ahead
always@(*)
  begin
    case(state)
//****************************************************************************
      S_LOAD_PC:  //read the new program counter off the data bus
        begin
          case(tmp_a[1:0])
            2'h0 :         next_pc = cpu_dar0;
            2'h1 :         next_pc = cpu_dar1;
            2'h2 :         next_pc = cpu_dar2;
            2'h3 :         next_pc = cpu_dar3;
          endcase
        end
//****************************************************************************
      S_EXECUTE_INST:
        begin
                           next_pc = pc_p1;
        end
//****************************************************************************
      S_EXECUTE_INST2:
        begin
          if(branch_taken) next_pc = branch;
          else             next_pc = pc_p2;
       end
//****************************************************************************
      S_EXECUTE_INST3:
        begin
          if(jsr_a|jmp_a) next_pc = add;
          else            next_pc = pc_p3;
        end
//****************************************************************************
      S_RTS2, S_RTI:      next_pc = tmp_pc;
//****************************************************************************
      default:            next_pc = pc;
    endcase
  end

//****************************************************************************
//****************************************************************************
//next if look-ahead
always@(*)
  begin
    case(state)
//****************************************************************************
      S_LOAD_PC:
        begin
          case(tmp_a[1:0])
            2'h0 :                   next_if = cpu_dar0[15:1] + 15'h0002;
            2'h1 :                   next_if = cpu_dar1[15:1] + 15'h0002;
            2'h2 :                   next_if = cpu_dar2[15:1] + 15'h0002;
            2'h3 :                   next_if = cpu_dar3[15:1] + 15'h0002;
          endcase
        end
//****************************************************************************
      S_INST_DELAY:                  next_if = cpu_ifp4;
//****************************************************************************
      S_LOAD_INST:                   next_if = cpu_ifp4;
//****************************************************************************
      S_EXECUTE_INST:
        begin
          if(ifReady)
            begin
              if(ifCycle)
                begin
                  if(iCount < 4'h2)  next_if = cpu_ifp4;
                  else               next_if = cpu_if;
                end
              else if(iCount < 4'h6) next_if = cpu_ifp4;
              else                   next_if = cpu_if;
            end
          else if(ifCycle)
            begin
              if(iCount < 4'h6)      next_if = cpu_ifp4;
              else                   next_if = cpu_if;
            end
          else if(iCount < 4'hA)     next_if = cpu_ifp4;
          else                       next_if = cpu_if;
        end
//****************************************************************************
      S_EXECUTE_INST2:
        begin
          if(branch_taken)           next_if = branch[15:1] + 15'h0002;
          else if(ifReady)
            begin
              if(ifCycle)
                begin
                  if(iCount < 4'h3)  next_if = cpu_ifp4;
                  else               next_if = cpu_if;
                end
              else if(iCount < 4'h7) next_if = cpu_ifp4;
              else                   next_if = cpu_if;
            end
          else if(ifCycle)
            begin
              if(iCount < 4'h7)      next_if = cpu_ifp4;
              else                   next_if = cpu_if;
            end
          else if(iCount < 4'hB)     next_if = cpu_ifp4;
          else                       next_if = cpu_if;
        end
//****************************************************************************
      S_EXECUTE_INST3:
        begin
          if(jsr_a|jmp_a)            next_if = add[15:1] + 15'h0002;
          else if(~st_axy)           next_if = cpu_if;
          else if(ifReady)
            begin
              if(ifCycle)
                begin
                  if(iCount < 4'h4)  next_if = cpu_ifp4;
                  else               next_if = cpu_if;
                end
              else if(iCount < 4'h8) next_if = cpu_ifp4;
              else                   next_if = cpu_if;
            end
          else if(ifCycle)
            begin
              if(iCount < 4'h8)      next_if = cpu_ifp4;
              else                   next_if = cpu_if;
            end
          else if(iCount < 4'hC)     next_if = cpu_ifp4;
          else                       next_if = cpu_if;
        end
//****************************************************************************
      S_IND_Y,S_REG_OP,S_ZP_REG_OP,S_BCD_DELAY,S_MEM_DELAY,S_ZP_MEM_DELAY:
        begin
          if(ifReady)
            begin
              if(ifCycle)
                begin
                  if(iCount == 4'h0)  next_if = cpu_ifp4;
                  else                next_if = cpu_if;
                end
              else if(iCount <= 4'h4) next_if = cpu_ifp4;
              else                    next_if = cpu_if;
            end
          else if(ifCycle)
            begin
              if(iCount <= 4'h4)      next_if = cpu_ifp4;
              else                    next_if = cpu_if;
            end
          else if(iCount <= 4'h8)     next_if = cpu_ifp4;
          else                        next_if = cpu_if;
        end
//****************************************************************************
      S_RTS2, S_RTI:                  next_if = tmp_pc[15:1] + 15'h0002;
//****************************************************************************
      S_IR_W_HIT:                     next_if = pc[15:1] + 15'h0002;
//****************************************************************************
      default:                        next_if = cpu_if;
    endcase
end
 
//****************************************************************************
//****************************************************************************
//next if look-ahead
always@(*)
  begin
    case(state)
//****************************************************************************
      S_EXECUTE_INST:
        begin
          if(brk)          next_sp = spm3;
          else if(rts)     next_sp = spp2;
          else if(rti)     next_sp = spp3;
          else if(txs)     next_sp = x;
          else if(plp|pla) next_sp = spp1;
          else if(pha|php) next_sp = spm1;
          else             next_sp = sp;
        end
//****************************************************************************
      S_EXECUTE_INST3:
        begin
          if(jsr_a)        next_sp = spm2;
          else             next_sp = sp;
       end
//****************************************************************************
      S_NMI,S_IRQ:         next_sp = spm3;
//****************************************************************************
      default:             next_sp = sp;
    endcase
  end
  
`ifdef debug
//debug trace registers
reg [959:0] pc_trace = 960'd0;
reg [959:0] ad_trace = 960'd0;
reg [959:0] aw_trace = 960'd0;
reg [15:2] cpu_al_i;
`endif

//synchronous cpu state machine
//****************************************************************************
//****************************************************************************
//instruction execution state machine controller
always@(posedge clk)
  begin

    //bus master logic
    vic_ack <= vic_req;
    dma_ack <= ~vic_req|dma_req|~dma_ack; //must only be a one cycle steal
    cpu_ack <= ~(vic_req & ~(vic_req & dma_ack & ~dma_req) & cpu_pause & resetIn & (&halt));    
    cpu_ack2 <= cpu_ack;
    
    if(~cpu_ack2)
      begin
        //synchronize the data buses
        if(pRead) cpu_dbr <= cpu_db;
        else
          case(opSelect)
            2'h0 : cpu_dbr <= cpu_ib[7:0];
            2'h1 : cpu_dbr <= cpu_ib[15:8];
            2'h2 : cpu_dbr <= cpu_ib[23:16];
            2'h3 : cpu_dbr <= cpu_ib[31:24];
          endcase
        cpu_ibr <= cpu_ib;
      end
       
    //cpu is being reset
    if(~resetIn)
      begin

`ifdef debug
        pc_trace <= 960'd0;
        ad_trace <= 960'd0;
        aw_trace <= 960'd0;
        breakTrig <= FALSE;
        breakTrig1 <= FALSE;
`endif
        
        halt <= 4'hF; 

        rnmi <= H;
        rnmi2 <= H; //this needs to be cleared to allow nmi's
        nmi <= H;
        rirq <= H;
        
        state <= S_RESET;

//the rest of these are optional
        ifCycle <= FALSE;
        ifReady <= FALSE;
        iCount <= 0;
        stackit <= 0;
      end
    else if(~cpu_ack) //cpu is the master
      begin
      
        opSelect <= cpu_a[1:0];
        
        state <= next_state;
        x <= next_x;
        y <= next_y;

        addx <= tmp_ax;
        addy <= { ah, imm_op } + next_y;
        pc <= next_pc;
        pc_p2 <= pcp2;
        cpu_if <= next_if;
        sp <= next_sp;
        spp1 <= next_spp1;
        spp2 <= next_spp2;
        spp3 <= next_spp3;
        
        if(state == S_EXECUTE_INST)       branch <= pc + 2'h3 + { {9{imm_op[7]}}, imm_op[6:0] };
        else if(state == S_EXECUTE_INST2) branch <= pc + 3'h4 + { {9{imm_op[7]}}, imm_op[6:0] };
        else if(state == S_EXECUTE_INST3) branch <= pc + 3'h5 + { {9{imm_op[7]}}, imm_op[6:0] };
        else                              branch <= pc + 2'h2 + { {9{imm_op[7]}}, imm_op[6:0] };

        cachei_r <= cache;
        
        if((next_state == S_EXECUTE_INST2) | (next_state == S_EXECUTE_INST3)) mode <= ira[7];
        
        //load #, zp cache operand or memory operand
        if(next_state == S_EXECUTE_INST2) alu_op <= imm_op;
        else if(next_state == S_BCD_DELAY) alu_op <= alu_op;
        else if((next_state == S_ZP_MEM_OP) | (next_state == S_ZP_REG_OP) | (next_state == S_ZP_MEM_DELAY))
          begin
            case(cpu_aw[1:0])
              0: alu_op <= cache[7:0];
              1: alu_op <= cache[15:8];
              2: alu_op <= cache[23:16];
              3: alu_op <= cache[31:24];
            endcase
          end
        else if(state == S_ZP_MEM_DELAY) alu_op <= alu_op;
        else if(cpu_ack2) alu_op <= cpu_dbr;
        else if(pRead) alu_op <= cpu_db;
        else
          case(opSelect)
            2'h0 : alu_op <= cpu_ib[7:0];
            2'h1 : alu_op <= cpu_ib[15:8];
            2'h2 : alu_op <= cpu_ib[23:16];
            2'h3 : alu_op <= cpu_ib[31:24];
          endcase
   
        cpu_ir <= cpu_ack2 ? cpu_ibr : cpu_ib;
        
        cpu_dar0 <= cpu_ack2 ? cpu_ibr[15:0]  : cpu_ib[15:0];
        cpu_dar1 <= cpu_ack2 ? cpu_ibr[23:8]  : cpu_ib[23:8];
        cpu_dar2 <= cpu_ack2 ? cpu_ibr[31:16] : cpu_ib[31:16];
        cpu_dar3 <= cpu_ack2 ? { cpu_ibr[7:0], cpu_ibr[31:24] } : { cpu_ib[7:0], cpu_ib[31:24] };

        ib0 <= cpu_ack2 ? (cpu_if[1] ? cpu_ibr[23:16] : cpu_ibr[7:0])   : (cpu_if[1] ? cpu_ib[23:16] : cpu_ib[7:0]);
        ib1 <= cpu_ack2 ? (cpu_if[1] ? cpu_ibr[31:24] : cpu_ibr[15:8])  : (cpu_if[1] ? cpu_ib[31:24] : cpu_ib[15:8]);
        ib2 <= cpu_ack2 ? (cpu_if[1] ? cpu_ibr[7:0]   : cpu_ibr[23:16]) : (cpu_if[1] ? cpu_ib[7:0]   : cpu_ib[23:16]);
        ib3 <= cpu_ack2 ? (cpu_if[1] ? cpu_ibr[15:8]  : cpu_ibr[31:24]) : (cpu_if[1] ? cpu_ib[15:8]  : cpu_ib[31:24]);

`ifdef debug
        breakTrig <= breakTrig1;
        ad_trace[15:0] <= cpu_a;
        ad_trace[959:16] <= ad_trace[943:0];
        aw_trace[15:0] <= cpu_aw;
        aw_trace[959:16] <= aw_trace[943:0];
        cpu_a2_i <= cpu_a;
        cpu_al_i <= cpu_al;
        cpu_aw2_i <= cpu_aw;
        cpu_drr <= cpu_ir;
        cpu_byte_enable <= cpu_wbyte;
        cpu_wb_r <= cpu_wb;

        //count events
        if(state == S_RESET) cycleCount <= 0; //count events
        //else cycleCount <= cycleCount + H;
        cpu_rw_r <= cpu_rw;
`endif

        ifReady <= ifCycle;
          
        //update interrupts
        rnmi <= nmi_req;
        rnmi2 <= rnmi;
        rirq <= irq | interrupt;
        
        if(~rnmi & rnmi2) nmi <= L; //edge trigger until processed
        
        case(state)
//****************************************************************************
          S_RESET:  //accessing reset vector (pc) from memory
            begin
              tmp_a <= V_RST;//reset vector
              ifCycle <= FALSE;
            end
//****************************************************************************
          S_PC_DELAY: ifCycle <= FALSE; //accessing vector or indirect jump pc from memory
//****************************************************************************
          S_LOAD_PC:  ifCycle <= TRUE; //read the new program counter off the data bus
//****************************************************************************
//accessing new instruction from memory, align cpu_if on a word boundary
//to make future instruction loading simpler...
          S_INST_DELAY:
            begin
              ifCycle <= TRUE;
              iCount <= 0;
              ir[95:0] <= 0;
            end
//****************************************************************************
//transfer the first instruction from the new path to cpu_ir & start accessing
//the next 4 bytes in the instruction stream
          S_LOAD_INST:
            begin
              case(pc[1:0])
                2'h0 :
                  begin
                    ir[31:0] <= cpu_ir;
                    iCount <= 4'h4;
                  end
                2'h1 :
                  begin
                    ir[23:0] <= cpu_ir[31:8];
                    iCount <= 4'h3;
                  end
                2'h2 :
                  begin
                    ir[15:0] <= cpu_ir[31:16];
                    ir[31:16] <= cpu_ir[15:0];
                    iCount <= 4'h4;
                  end
                2'h3 :
                  begin
                    ir[7:0] <= cpu_ir[31:24];
                    ir[23:8] <= cpu_ir[15:0];
                    iCount <= 4'h3;
                  end
              endcase

              ifCycle <= TRUE;
            end

//****************************************************************************
//start executing a single byte instruction
          S_EXECUTE_INST:
            begin
`ifdef debug
              breakTrig1 <= (pc == break_pc);
              pc_trace[15:0] <= pc;
              pc_trace[959:16] <= pc_trace[943:0];
`endif

//remember operations for later
              sta_ix_2 <= sta_ix;
              sta_iy_2 <= sta_iy;
              
              asl <= asl_z|asl_zx|asl_a|asl_ax;
              rol <= rol_z|rol_zx|rol_a|rol_ax;
              lsr <= lsr_z|lsr_zx|lsr_a|lsr_ax;
              ror <= ror_z|ror_zx|ror_a|ror_ax;
              dec <= dec_z|dec_zx|dec_a|dec_ax;
              inc <= inc_z|inc_zx|inc_a|inc_ax;

              ora <= ora_iy|ora_ix|ora_z|ora_zx|ora_a|ora_ay|ora_ax;
              bita <= bita_z|bita_a;
              anda <= anda_iy|anda_ix|anda_z|anda_zx|anda_a|anda_ay|anda_ax;
              eora <= eora_iy|eora_ix|eora_z|eora_zx|eora_a|eora_ay|eora_ax;
              
              addsub <= adca_z|adca_zx|adca_ix|adca_iy|sbca_z|sbca_zx|sbca_ix|sbca_iy|adca_a|adca_ax|adca_ay|sbca_a|sbca_ax|sbca_ay;
              
              cmpa_2 <= cmpa_iy|cmpa_ix|cmpa_z|cmpa_zx|cmpa_z|cmpa_zx|cmpa_a|cmpa_ay|cmpa_ax;
              cpx_2 <= cpx_z|cpx_a;
              cpy_2 <= cpy_z|cpy_a;
              
              ldx <= ldx_zy|ldx_z|ldx_a|ldx_ay;
              lda <= lda_iy|lda_ix|lda_z|lda_zx|lda_a|lda_ay|lda_ax|pla;
              ldy <= ldy_z|ldy_zx|ldy_a|ldy_ax;

              sty <= sty_a|sty_z|sty_zx;
              stx <= stx_a|stx_z|stx_zy;
              
              mem_op_2 <= mem_op;
              
              stackit <= 0; //stacking stuff for an interrupt, brk or jsr
              
///deal with program counter change possibilities first...
              if(brk)
                begin
                  stackit <= { H, pc[15:0], sp_brk };
//brk is same as irq except for break status bit set on the stack and pc + 2
//is stacked, note this is a software interrupt, uses irq vector
                  interrupt <= H;
                  tmp_a <= V_IRQ;  
                end
              else if(rts)
                begin
                  cacheiSelect <= sp[1:0];
                end
              else if(rti)
                begin
                  cacheiSelect <= sp[1:0];
                  case(sp[1:0])
                    0: tmp_pc <= caches[31:16];
                    1: tmp_pc <= { caches[7:0], caches[31:24] };
                    2: tmp_pc <= caches[15:0];
                    3: tmp_pc <= caches[23:8];
                  endcase
                end
//****************************************************************************
// 1 byte instructions
              else
                begin
                  if(ifReady)
                    begin
                      //subtracting 1 and adding 4
                      iCount <= iCount + 4'h3;
                      case(iCount)
                      
                        4'h9:
                          begin
                            ir[63:0] <= ir[71:8];
                            ir[95:64] <= ib32;
                          end
                        4'h8:
                          begin
                            ir[55:0] <= ir[63:8];
                            ir[87:56] <= ib32;
                          end
                        4'h7:
                          begin
                            ir[47:0] <= ir[55:8];
                            ir[79:48] <= ib32;
                          end
                        4'h6:
                          begin
                            ir[39:0] <= ir[47:8];
                            ir[71:40] <= ib32;
                          end
                        4'h5:
                          begin
                            ir[31:0] <= ir[39:8];
                            ir[63:32] <= ib32;
                          end
                        4'h4:
                          begin
                            ir[23:0] <= ir[31:8];
                            ir[55:24] <= ib32;
                          end
                        4'h3:
                          begin
                            ir[15:0] <= ir[23:8];
                            ir[47:16] <= ib32;
                          end
                        4'h2:
                          begin
                            ir[7:0] <= ir[15:8];
                            ir[39:8]  <= ib32;
                          end
                        4'h1: ir[31:0]  <= ib32;

                        //this should not happen... 
                        default: halt <= 4'h0; 
                      endcase
                    end
                  else
                    begin
                      iCount <= iCount - 4'h1;
                      ir[95:88] <= ZERO;
                      ir[87:0] <= ir[95:8];
                    end
                    
                  if(ifReady)
                    begin
                      if(ifCycle)
                        begin
                          if(iCount >= 4'h2) ifCycle <= FALSE;
                        end
                      else if(iCount < 4'h6) ifCycle <= TRUE;
                    end
                  else if(ifCycle)
                    begin
                      if(iCount >= 4'h6) ifCycle <= FALSE;
                    end
                  else if(iCount < 4'hA) ifCycle <= TRUE;
                
//****************************************************************************

//                  if(txs) //nothing to be done
                  if(plp)
                    begin
                      case(sp[1:0])
                        0:
                          begin
                            carry <= cache[8];
                            zero <= cache[9];
                            interrupt <= cache[10];
                            decimal <= cache[11];
                            overflow <= cache[14];
                            negative <= cache[15];
                          end
                        1:
                          begin
                            carry <= cache[16];
                            zero <= cache[17];
                            interrupt <= cache[18];
                            decimal <= cache[19];
                            overflow <= cache[22];
                            negative <= cache[23];
                          end
                        2:
                          begin
                            carry <= cache[24];
                            zero <= cache[25];
                            interrupt <= cache[26];
                            decimal <= cache[27];
                            overflow <= cache[30];
                            negative <= cache[31];
                          end
                        3:
                          begin
                            carry <= cache[0];
                            zero <= cache[1];
                            interrupt <= cache[2];
                            decimal <= cache[3];
                            overflow <= cache[6];
                            negative <= cache[7];
                          end
                      endcase
                    end

//                  if(pha|php) //nothing to be done
                  if(pla)
                    begin
                      tmp_a <= spp1;
                    end
                  if(asla)
                    begin
                      carry <= a[7];
                      negative <= a[6];
                      zero <= ~(|a[6:0]);
                      a <= { a[6:0], L };
                    end
                  if(rola)
                    begin
                      carry <= a[7];
                      negative <= a[6];
                      zero <=  ~carry & ~(|a[6:0]);
                      a <= { a[6:0], carry };
                    end
                  if(lsra)
                    begin
                      carry <= a[0];
                      negative <= L;
                      zero <= ~(|a[7:1]);
                      a <= { L, a[7:1] };
                    end
                  if(rora)
                    begin
                      carry <= a[0];
                      negative <= carry;
                      zero <= ~carry & ~(|a[7:1]);
                      a <= { carry, a[7:1] };
                    end
                  if(tya)
                    begin
                      zero <= (y == ZERO);
                      negative <= y[7];
                      a <= y;
                    end
                  if(txa)
                    begin
                      zero <= (x == ZERO);
                      negative <= x[7];
                      a <= x;
                    end

                  if(tay)
                    begin
                      zero <= (a == ZERO);
                      negative <= a[7];
                    end
                  if(dey)
                    begin
                      zero <= (y == ONE);
                      negative <= (y == ZERO) ? H : (y == 8'h80) ? L : y[7];
                    end
                  if(iny)
                    begin
                      zero <= (y == NEG_ONE);
                      negative <= (y == NEG_ONE) ? L : (y == 8'h7F) ? H : y[7];
                    end

                  if(tax)
                    begin
                      zero <= (a == ZERO);
                      negative <= a[7];
                    end
                  if(tsx)
                    begin
                      zero <= (sp == ZERO);
                      negative <= sp[7];
                    end
                  if(dex)
                    begin
                      zero <= (x == ONE);
                      negative <= (x == ZERO) ? H : (x == 8'h80) ? L : x[7];
                    end
                  if(inx)
                    begin
                      zero <= (x == NEG_ONE);
                      negative <= (x == NEG_ONE) ? H : (x == 8'h7F) ? H : x[7];
                    end

//                if(nop) //nothing to be done
                  if(cli) interrupt <= FALSE;
                  if(sei) interrupt <= TRUE;
                  if(clc) carry <= FALSE;
                  if(sec) carry <= TRUE;
                  if(cld) decimal <= FALSE;
                  if(sed) decimal <= TRUE;
                  if(clv) overflow <= FALSE;
                end
            end
//****************************************************************************
//start executing a 2 byte instruction
          S_EXECUTE_INST2:
            begin

`ifdef debug
              breakTrig1 <= (pc == break_pc);
              pc_trace[15:0] <= pc;
              pc_trace[959:16] <= pc_trace[943:0];
`endif

//remember operations for later
              sta_ix_2 <= sta_ix;
              sta_iy_2 <= sta_iy;
              
              asl <= asl_z|asl_zx|asl_a|asl_ax;
              rol <= rol_z|rol_zx|rol_a|rol_ax;
              lsr <= lsr_z|lsr_zx|lsr_a|lsr_ax;
              ror <= ror_z|ror_zx|ror_a|ror_ax;
              dec <= dec_z|dec_zx|dec_a|dec_ax;
              inc <= inc_z|inc_zx|inc_a|inc_ax;

              ora <= ora_iy|ora_ix|ora_z|ora_zx|ora_a|ora_ay|ora_ax;
              bita <= bita_z|bita_a;
              anda <= anda_iy|anda_ix|anda_z|anda_zx|anda_a|anda_ay|anda_ax;
              eora <= eora_iy|eora_ix|eora_z|eora_zx|eora_a|eora_ay|eora_ax;
              
              addsub <= adca_z|adca_zx|adca_ix|adca_iy|sbca_z|sbca_zx|sbca_ix|sbca_iy|adca_a|adca_ax|adca_ay|sbca_a|sbca_ax|sbca_ay;
              
              cmpa_2 <= cmpa_iy|cmpa_ix|cmpa_z|cmpa_zx|cmpa_z|cmpa_zx|cmpa_a|cmpa_ay|cmpa_ax;
              cpx_2 <= cpx_z|cpx_a;
              cpy_2 <= cpy_z|cpy_a;
              
              ldx <= ldx_zy|ldx_z|ldx_a|ldx_ay;
              lda <= lda_iy|lda_ix|lda_z|lda_zx|lda_a|lda_ay|lda_ax|pla;
              ldy <= ldy_z|ldy_zx|ldy_a|ldy_ax;

              sty <= sty_a|sty_z|sty_zx;
              stx <= stx_a|stx_z|stx_zy;
              
              mem_op_2 <= mem_op;
              
//deal with program counter change possibilities first...
              if(branch_taken) stackit <= 0; //nothing to be done here 
              else if(two_byte) //necessary to catch illegal instructions below
                begin
                  if(ifReady)
                    begin
                      //subtracting 2 and adding 4
                      iCount <= iCount + 4'h2;
                      case(iCount)

                        4'hA:
                          begin
                            ir[63:0] <= ir[79:16];
                            ir[95:64] <= ib32;
                          end
                        4'h9:
                          begin
                            ir[55:0] <= ir[71:16];
                            ir[87:56] <= ib32;
                          end
                        4'h8:
                          begin
                            ir[47:0] <= ir[63:16];
                            ir[79:48] <= ib32;
                          end
                        4'h7:
                          begin
                            ir[39:0] <= ir[55:16];
                            ir[71:40] <= ib32;
                          end
                        4'h6:
                          begin
                            ir[31:0] <= ir[47:16];
                            ir[63:32] <= ib32;
                          end
                        4'h5:
                          begin
                            ir[23:0] <= ir[39:16];
                            ir[55:24] <= ib32;
                          end
                        4'h4:
                          begin
                            ir[15:0] <= ir[31:16];
                            ir[47:16] <= ib32;
                          end
                        4'h3:
                          begin
                            ir[7:0] <= ir[23:16];
                            ir[39:8]  <= ib32;
                          end
                        4'h2: ir[31:0]  <= ib32;

                        default: halt <= 4'h3; //this should not happen
                      endcase
                    end
                  else
                    begin
                      iCount <= iCount - 4'h2;
                      ir[95:80] <= 16'h0000;
                      ir[79:0] <= ir[95:16];
                    end

                  if(anda_i)
                    begin
                      zero <= and_op_zero;
                      negative <= and_op[7];
                      a <= and_op;
                    end
                  if(eora_i)
                    begin
                      zero <= eor_op_zero;
                      negative <= eor_op[7];
                      a <= eor_op;
                    end
                  if(ora_i)
                    begin
                      zero <= or_op_zero;
                      negative <= or_op[7];
                      a <= or_op;
                    end
                  if(lda_i)
                    begin
                      zero <= alu_op_0;
                      negative <= alu_op[7];
                      a <= alu_op;
                    end
                  if((adca_i|sbca_i) & ~decimal)
                    begin
                      zero <= is_zero_a;
                      carry <= Cout_a;
                      negative <= is_negative_a;
                      overflow <= is_overflow_a;
                      a <= result;
                    end
                  if(cmpa_i)
                    begin
                      zero <= is_zero_ca;
                      carry <= Cout_ca;
                      negative <= N_ca;
                    end

                  if(ldy_i)
                    begin
                      zero <= alu_op_0;
                      negative <= alu_op[7];
                    end
                  if(cpy_i)
                    begin
                      zero <= is_zero_y;
                      carry <= Cout_y;
                      negative <= N_y;
                    end

                  if(ldx_i)
                    begin
                      zero <= alu_op_0;
                      negative <= alu_op[7];
                    end
                  if(cpx_i)
                    begin
                      zero <= is_zero_x;
                      carry <= Cout_x;
                      negative <= N_x;
                    end
                  
                  if(ldx_zy|stx_zy) tmp_a <= { ZERO, addy[7:0] };
                  else if(zx)       tmp_a <= { ZERO, addx[7:0] };
                  else if(ix)
                    begin
                      case(addx[1:0])
                        2'h0 :      tmp_ixy <= cachez[15:0];
                        2'h1 :      tmp_ixy <= cachez[23:8];
                        2'h2 :      tmp_ixy <= cachez[31:16];
                        2'h3 :      tmp_ixy <= { cachez[7:0], cachez[31:24] };
                      endcase
                    end
                  else if(iy)       cacheiSelect <= add[1:0];
                  else              tmp_a <= { ZERO, add[7:0] };

                  if(ifReady)
                    begin
                      if(ifCycle)
                        begin
                          if(iCount >= 4'h3) ifCycle <= FALSE;
                        end
                      else if(iCount < 4'h7) ifCycle <= TRUE;
                    end
                  else if(ifCycle)
                    begin
                      if(iCount >= 4'h7) ifCycle <= FALSE;
                    end
                  else if(iCount < 4'hB) ifCycle <= TRUE;
                end
//****************************************************************************
//illegal instruction
              else halt <= 4'hE;
            end
//****************************************************************************
//start processing an nmi
          S_NMI:
            begin
              stackit <= { H, pc, sp_int };
//re-enable NMI, cannot become active until the first instruction of service routine
              nmi <= TRUE; 
              tmp_a <= V_NMI;
            end
//****************************************************************************
//start processing an irq
          S_IRQ:
            begin
              stackit <= { H, pc, sp_int };
              interrupt <= H;
              tmp_a <= V_IRQ;
            end
//****************************************************************************
//reload ir register
          S_IR_W_HIT:
            begin
              stackit <= 0;
`ifdef debug
              cycleCount <= cycleCount + H;
`endif
            end
//****************************************************************************
//start executing a 3 byte instruction
          S_EXECUTE_INST3:
            begin
`ifdef debug
              breakTrig1 <= (pc == break_pc);
              pc_trace[15:0] <= pc;
              pc_trace[959:16] <= pc_trace[943:0];
`endif

//remember operations for later
              sta_ix_2 <= sta_ix;
              sta_iy_2 <= sta_iy;
              
              asl <= asl_z|asl_zx|asl_a|asl_ax;
              rol <= rol_z|rol_zx|rol_a|rol_ax;
              lsr <= lsr_z|lsr_zx|lsr_a|lsr_ax;
              ror <= ror_z|ror_zx|ror_a|ror_ax;
              dec <= dec_z|dec_zx|dec_a|dec_ax;
              inc <= inc_z|inc_zx|inc_a|inc_ax;

              ora <= ora_iy|ora_ix|ora_z|ora_zx|ora_a|ora_ay|ora_ax;
              bita <= bita_z|bita_a;
              anda <= anda_iy|anda_ix|anda_z|anda_zx|anda_a|anda_ay|anda_ax;
              eora <= eora_iy|eora_ix|eora_z|eora_zx|eora_a|eora_ay|eora_ax;
              
              addsub <= adca_z|adca_zx|adca_ix|adca_iy|sbca_z|sbca_zx|sbca_ix|sbca_iy|adca_a|adca_ax|adca_ay|sbca_a|sbca_ax|sbca_ay;
              
              cmpa_2 <= cmpa_iy|cmpa_ix|cmpa_z|cmpa_zx|cmpa_z|cmpa_zx|cmpa_a|cmpa_ay|cmpa_ax;
              cpx_2 <= cpx_z|cpx_a;
              cpy_2 <= cpy_z|cpy_a;
              
              ldx <= ldx_zy|ldx_z|ldx_a|ldx_ay;
              lda <= lda_iy|lda_ix|lda_z|lda_zx|lda_a|lda_ay|lda_ax|pla;
              ldy <= ldy_z|ldy_zx|ldy_a|ldy_ax;

              sty <= sty_a|sty_z|sty_zx;
              stx <= stx_a|stx_z|stx_zy;
              
              mem_op_2 <= mem_op;
              
              //stacking stuff for an interrupt, brk or jsr
              stackit <= 0;
              
//deal with program counter change possibilities first...
              if(jsr_a)      stackit <= { H, ZERO, pc_p2 };
              else if(jmp_a) stackit <= 1; //nothing to be done here
              else if(jmp_i) tmp_a <= add;
              else
                begin
                  if(ifReady)
                    begin
                      //subtracting 3 and adding 4
                      iCount <= iCount + 4'h1;
                      case(iCount)

                        4'hB:
                          begin
                            ir[63:0] <= ir[87:24];
                            ir[95:64] <= ib32;
                          end
                        4'hA:
                          begin
                            ir[55:0] <= ir[79:24];
                            ir[87:56] <= ib32;
                          end
                        4'h9:
                          begin
                            ir[47:0] <= ir[71:24];
                            ir[79:48] <= ib32;
                          end
                        4'h8:
                          begin
                            ir[39:0] <= ir[63:24];
                            ir[71:40] <= ib32;
                          end
                        4'h7:
                          begin
                            ir[31:0] <= ir[55:24];
                            ir[63:32] <= ib32;
                          end
                        4'h6:
                          begin
                            ir[23:0] <= ir[47:24];
                            ir[55:24] <= ib32;
                          end
                        4'h5:
                          begin
                            ir[15:0] <= ir[39:24];
                            ir[47:16] <= ib32;
                          end
                        4'h4:
                          begin
                            ir[7:0] <= ir[31:24];
                            ir[39:8]  <= ib32;
                          end

                        4'h3: ir[31:0]  <= ib32;

                        default: halt <= 4'h2; //this should not happen 
                      endcase
                    end
                  else
                    begin
                      iCount <= iCount - 4'h3;
                      ir[95:72] <= 24'h000000;
                      ir[71:0] <= ir[95:24];
                    end

                  if(~st_axy)                ifCycle <= FALSE;
                  else if(ifReady)
                    begin
                      if(ifCycle)
                        begin
                          if(iCount >= 4'h4) ifCycle <= FALSE;
                        end
                      else if(iCount < 4'h8) ifCycle <= TRUE;
                    end
                  else if(ifCycle)
                    begin
                      if(iCount >= 4'h8)     ifCycle <= FALSE;
                    end
                  else if(iCount < 4'hC)     ifCycle <= TRUE;

                  if(ax)      tmp_a <= addx;
                  else if(ay) tmp_a <= addy;
                  else        tmp_a <= add;
                end
            end
//****************************************************************************
          S_IND_X:  // indirect x
            begin
              tmp_a <= tmp_ixy;
              
              if(ifReady) //adding 4
                begin
                  iCount <= iCount + 4'h4;
                  case(iCount)
                    4'h8: ir[95:64] <= ib32;
                    4'h7: ir[87:56] <= ib32;
                    4'h6: ir[79:48] <= ib32;
                    4'h5: ir[71:40] <= ib32;
                    4'h4: ir[63:32] <= ib32;
                    4'h3: ir[55:24] <= ib32;
                    4'h2: ir[47:16] <= ib32;
                    4'h1: ir[39:8]  <= ib32;
                    4'h0: ir[31:0]  <= ib32;
                    default: halt <= 4'h7; //this should not happen
                  endcase
                end
                ifCycle <= FALSE;
            end
//****************************************************************************
          S_IND_Y:  // indirect y 
            begin
              case(cacheiSelect)
                2'h0 : tmp_ixy <= cachei_r[15:0] + y;
                2'h1 : tmp_ixy <= cachei_r[23:8] + y;
                2'h2 : tmp_ixy <= cachei_r[31:16] + y;
                2'h3 : tmp_ixy <= { cachei_r[7:0], cachei_r[31:24] } + y;
              endcase

              if(ifReady) //adding 4
                begin
                  iCount <= iCount + 4'h4;
                  case(iCount)
                    4'h8: ir[95:64] <= ib32;
                    4'h7: ir[87:56] <= ib32;
                    4'h6: ir[79:48] <= ib32;
                    4'h5: ir[71:40] <= ib32;
                    4'h4: ir[63:32] <= ib32;
                    4'h3: ir[55:24] <= ib32;
                    4'h2: ir[47:16] <= ib32;
                    4'h1: ir[39:8]  <= ib32;
                    4'h0: ir[31:0]  <= ib32;
                    default: halt <= 4'h7; //this should not happen
                  endcase
                end

              if(ifReady)
                begin
                  if(ifCycle)
                    begin
                      if(iCount > 4'h0) ifCycle <= FALSE;
                    end
                  else if(iCount <= 4'h4) ifCycle <= TRUE;
                end
              else if(ifCycle)
                begin
                  if(iCount > 4'h4) ifCycle <= FALSE;
                end
              else if(iCount <= 4'h8) ifCycle <= TRUE;
            end
//****************************************************************************
          S_IND_Y2:  // indirect y 
            begin
              tmp_a <= tmp_ixy;

              if(ifReady) //adding 4
                begin
                  iCount <= iCount + 4'h4;
                  case(iCount)
                    4'h8: ir[95:64] <= ib32;
                    4'h7: ir[87:56] <= ib32;
                    4'h6: ir[79:48] <= ib32;
                    4'h5: ir[71:40] <= ib32;
                    4'h4: ir[63:32] <= ib32;
                    4'h3: ir[55:24] <= ib32;
                    4'h2: ir[47:16] <= ib32;
                    4'h1: ir[39:8]  <= ib32;
                    4'h0: ir[31:0]  <= ib32;
                    default: halt <= 4'h7; //this should not happen
                  endcase
                end
              ifCycle <= FALSE;
            end
//****************************************************************************
          S_MEM_DELAY,S_ZP_MEM_DELAY: // reading memory operand
            begin
              if(ifReady) //adding 4
                begin
                  iCount <= iCount + 4'h4;
                  case(iCount)
                    4'h8: ir[95:64] <= ib32;
                    4'h7: ir[87:56] <= ib32;
                    4'h6: ir[79:48] <= ib32;
                    4'h5: ir[71:40] <= ib32;
                    4'h4: ir[63:32] <= ib32;
                    4'h3: ir[55:24] <= ib32;
                    4'h2: ir[47:16] <= ib32;
                    4'h1: ir[39:8]  <= ib32;
                    4'h0: ir[31:0]  <= ib32;
                    default: halt <= 4'h6; //this should not happen
                  endcase
                end

              if(ifReady)
                begin
                  if(ifCycle)
                    begin
                      if(iCount > 4'h0) ifCycle <= FALSE;
                    end
                  else if(iCount <= 4'h4) ifCycle <= TRUE;
                end
              else if(ifCycle)
                begin
                  if(iCount > 4'h4) ifCycle <= FALSE;
                end
              else if(iCount <= 4'h8) ifCycle <= TRUE;
            end
//****************************************************************************
          S_MEM_OP,S_ZP_MEM_OP: // complete memory operation
            begin
              if(ifReady) //adding 4
                begin
                  iCount <= iCount + 4'h4;
                  case(iCount)
                    4'h8: ir[95:64] <= ib32;
                    4'h7: ir[87:56] <= ib32;
                    4'h6: ir[79:48] <= ib32;
                    4'h5: ir[71:40] <= ib32;
                    4'h4: ir[63:32] <= ib32;
                    4'h3: ir[55:24] <= ib32;
                    4'h2: ir[47:16] <= ib32;
                    4'h1: ir[39:8]  <= ib32;
                    4'h0: ir[31:0]  <= ib32;
                    default: halt <= 4'h8; //this should not happen
                  endcase
                end
              ifCycle <= FALSE;

              if(asl)
                begin
                  carry <= alu_op[7];
                  negative <= alu_op[6];
                  zero <= ~(|alu_op[6:0]);
                end
              if(rol)
                begin
                  carry <= alu_op[7];
                  negative <= alu_op[6];
                  zero <= ~carry & ~(|alu_op[6:0]);
                end
              if(lsr)
                begin
                  carry <= alu_op[0];
                  negative <= L;
                  zero <= ~(|alu_op[7:1]);
                end
              if(ror)
                begin
                  carry <= alu_op[0];
                  negative <= carry;
                  zero <= ~carry & ~(|alu_op[7:1]);
                end
              if(dec)
                begin
                  zero <= alu_op_1;
                  negative <= alu_op_dec_neg;
                 end
              if(inc)
                begin
                  zero <= alu_op_255;
                  negative <= alu_op_inc_neg;
                end
            end
//****************************************************************************
          S_REG_OP, S_ZP_REG_OP: // complete register operation
            begin
              if(ifReady) //adding 4
                begin
                  iCount <= iCount + 4'h4;
                  case(iCount)
                    4'h8: ir[95:64] <= ib32;
                    4'h7: ir[87:56] <= ib32;
                    4'h6: ir[79:48] <= ib32;
                    4'h5: ir[71:40] <= ib32;
                    4'h4: ir[63:32] <= ib32;
                    4'h3: ir[55:24] <= ib32;
                    4'h2: ir[47:16] <= ib32;
                    4'h1: ir[39:8]  <= ib32;
                    4'h0: ir[31:0]  <= ib32;
                    default: halt <= 4'h9; //this should not happen
                  endcase
                end
              
              if(ifReady)
                begin
                  if(ifCycle)
                    begin
                      if(iCount > 4'h0) ifCycle <= FALSE;
                    end
                  else if(iCount <= 4'h4) ifCycle <= TRUE;
                end
              else if(ifCycle)
                begin
                  if(iCount > 4'h4) ifCycle <= FALSE;
                end
              else if(iCount <= 4'h8) ifCycle <= TRUE;

              if(ora)
                begin
                  zero <= or_op_zero;
                  negative <= or_op[7];
                  a <= or_op;
                end
              if(bita)
                begin
                  zero <= and_op_zero;
                  negative <= alu_op[7];
                  overflow <= alu_op[6];
                end
              if(anda)
                begin
                  zero <= and_op_zero;
                  negative <= and_op[7];
                  a <= and_op;
                end
              if(eora)
                begin
                  zero <= eor_op_zero;
                  negative <= eor_op[7];
                  a <= eor_op;
                end
              if(cmpa_2)
                begin
                  zero <= is_zero_ca;
                  carry <= Cout_ca;
                  negative <= N_ca;
                end
              if(cpy_2)
                begin
                  zero <= is_zero_y;
                  carry <= Cout_y;
                  negative <= N_y;
                end
              if(cpx_2)
                begin
                  zero <= is_zero_x;
                  carry <= Cout_x;
                  negative <= N_x;
                end
              if(ldx)
                begin
                  zero <= alu_op_0;
                  negative <= alu_op[7];
                end
              if(lda)
                begin
                  zero <= alu_op_0;
                  negative <= alu_op[7];
                  a <= alu_op;
                end
              if(ldy)
                begin
                  zero <= alu_op_0;
                  negative <= alu_op[7];
                end
              if(addsub & ~decimal)
                begin
                  a <= result;
                  zero <= is_zero_a;
                  carry <= Cout_a;
                  overflow <= is_overflow_a;
                  negative <= is_negative_a;
                end
            end
//****************************************************************************
          S_RTS: //rts
            begin
              ifCycle <= FALSE;
              case(cacheiSelect)
                0: tmp_pc <= cachei_r[23:8] + H;
                1: tmp_pc <= cachei_r[31:16] + H;
                2: tmp_pc <= { cachei_r[7:0], cachei_r[31:24] } + H;
                3: tmp_pc <= cachei_r[15:0] + H;
              endcase
            end
//****************************************************************************
          S_RTS2: ifCycle <= TRUE;
//****************************************************************************
          S_RTI: // rti
            begin
              ifCycle <= TRUE;
              case(cacheiSelect)
                0:
                  begin
                    carry <= cachei_r[8];
                    zero <= cachei_r[9];
                    interrupt <= cachei_r[10];
                    decimal <= cachei_r[11];
                    overflow <= cachei_r[14];
                    negative <= cachei_r[15];
                  end
                1:
                  begin
                    carry <= cachei_r[16];
                    zero <= cachei_r[17];
                    interrupt <= cachei_r[18];
                    decimal <= cachei_r[19];
                    overflow <= cachei_r[22];
                    negative <= cachei_r[23];
                  end
                2:
                  begin
                    carry <= cachei_r[24];
                    zero <= cachei_r[25];
                    interrupt <= cachei_r[26];
                    decimal <= cachei_r[27];
                    overflow <= cachei_r[30];
                    negative <= cachei_r[31];
                 end
                3:
                  begin
                    carry <= cachei_r[0];
                    zero <= cachei_r[1];
                    interrupt <= cachei_r[2];
                    decimal <= cachei_r[3];
                    overflow <= cachei_r[6];
                    negative <= cachei_r[7];
                  end
              endcase
            end
//****************************************************************************
          S_BCD_DELAY: //bcd arithmatic delay
            begin
              if(ifReady) //adding 4
                begin
                  iCount <= iCount + 4'h4;
                  case(iCount)
                    4'h8: ir[95:64] <= ib32;
                    4'h7: ir[87:56] <= ib32;
                    4'h6: ir[79:48] <= ib32;
                    4'h5: ir[71:40] <= ib32;
                    4'h4: ir[63:32] <= ib32;
                    4'h3: ir[55:24] <= ib32;
                    4'h2: ir[47:16] <= ib32;
                    4'h1: ir[39:8]  <= ib32;
                    4'h0: ir[31:0]  <= ib32;
                    default: halt <= 4'h9; //this should not happen
                  endcase
                end

              if(ifReady)
                begin
                  if(ifCycle)
                    begin
                      if(iCount > 4'h0) ifCycle <= FALSE;
                    end
                  else if(iCount <= 4'h4) ifCycle <= TRUE;
                end
              else if(ifCycle)
                begin
                  if(iCount > 4'h4) ifCycle <= FALSE;
                end
              else if(iCount <= 4'h8) ifCycle <= TRUE;

              a <= bcdResult;
              zero <= bcdZero_a;
              carry <= bcdCout_a;
              overflow <= bcdOverflow_a;
              negative <= bcdNegative_a;
            end
//****************************************************************************
          S_WRITE: // complete memory write
            begin
              if(ifReady) //adding 4
                begin
                  iCount <= iCount + 4'h4;
                  case(iCount)
                    4'h8: ir[95:64] <= ib32;
                    4'h7: ir[87:56] <= ib32;
                    4'h6: ir[79:48] <= ib32;
                    4'h5: ir[71:40] <= ib32;
                    4'h4: ir[63:32] <= ib32;
                    4'h3: ir[55:24] <= ib32;
                    4'h2: ir[47:16] <= ib32;
                    4'h1: ir[39:8]  <= ib32;
                    4'h0: ir[31:0]  <= ib32;
                    default: halt <= 4'h8; //this should not happen
                  endcase
                end
              ifCycle <= FALSE;
            end
        endcase
    end
end

`ifdef pport

cpupport cpu_pport(clk
             ,result,a,x,y,sp
             ,pc
             ,cpu_if,ifCycle,ifReady,iCount,ir
             ,carry,zero,interrupt,decimal,overflow,negative,irq,nmi_req,p_Read,cpu_rw_r
             ,pp_add, pp_Dout//,EppDstb
             );

`endif

`ifdef debug

cpu_debug dbg(clk, pc_trace, ad_trace, aw_trace
             ,state,halt,cpu_rw_r,cpu_byte_enable,a1,negative,decimal,interrupt,overflow,zero,carry,rirq,nmi
             ,ir,iCount,ifCycle,ifReady
             ,a,x,y,sp,alu_op,ib0,ib1,ib2,ib3,alu_op,cpu_wb_r
             ,pc,tmp_a,cpu_if,cpu_a2_i,cpu_al_i,cpu_aw2_i,break_pc
             ,cpu_drr,cycleCount
             ,debug_add, debug_Dout, debug_sel
             );
`endif

endmodule


