//****************************************************************************
// C64 VIC.
//
// (C) 2013, Ken Chamberlain <kenneth.r.chamberlain@comcast.net>
//
// Feel free to use this code in any project (commercial or not), as long as you
// keep this message, and the copyright notice. This code is provided "as is", 
// without any warranties of any kind. 
// 
//****************************************************************************

//`define debug
//`define la
//`define mouser

module X6567( mclk, vclk, red, green, blue, VS, HS, reset
            , dAB, we, re, dr, dw
            , vic_a_out ,memory, vic_color, vic_req, charRom
            , va15, va14
            , LP
            , irq
`ifdef mouser
            , mouse_x, mouse_y
`endif            
`ifdef la
            , laCounter, la_sync, lain, samplingDone
`endif
`ifdef debug
            , debug_Address, debug_Data , debug_sel
`endif
            ); //each 6 bits represents a character - 8 characters per address.


input mclk, vclk, we, re, va15, va14, LP, reset;
input [31:0] memory;
input [15:0] vic_color;
input [7:0] dw;

`ifdef mouser
input [47:0] dAB;
input [8:0] mouse_x;
input [7:0] mouse_y;
`else
input [46:0] dAB;
`endif

`ifdef debug
input [47:0] debug_Data;
input [2:0] debug_sel;
output [4:0] debug_Address;
`endif
`ifdef la
input lain, samplingDone;
output [4:0] laCounter;
output la_sync;
`endif

output vic_req, charRom, VS, HS, irq;
output [15:1] vic_a_out;
output [7:0] dr;
output [2:0] green, blue;
output [1:0] red;

parameter H = 1'b1;
parameter L = 1'b0;
parameter trigger_dma = 11;//on hcount
parameter dma_Count = 65;//dma cycle count max

//vic default colors - done this way to avoid a block ram being wasted as a 16x8 rom
// weighting                43421421
// channels                 RRGGGBBB
parameter color_black  = 8'b00000000;
parameter color_white  = 8'b11111111;
parameter color_red    = 8'b11000000;
parameter color_cyan   = 8'b11101110;
parameter color_purple = 8'b10001110;
parameter color_green  = 8'b01101001;
parameter color_blue   = 8'b00000110;//8'b00000101;
parameter color_yellow = 8'b11111010;
parameter color_orange = 8'b11010000;
parameter color_brown  = 8'b01001000;
parameter color_Lred   = 8'b11011011;
parameter color_Dgray  = 8'b01011100;
parameter color_Mgray  = 8'b01011011;
parameter color_Lgreen = 8'b10111100;
parameter color_Lblue  = 8'b01011111;
parameter color_Lgray  = 8'b10101101;

//#################################################
//vic character buffer - a character and its color
//ram are only loaded once, after that the data is
//recycled 8 times for each active raster - this is
//the way the real vic chip works - this does not
//allow for video splits in the middle of a character
//very well...Characters are loaded on the previous
//raster line before being displayed - the vertical
//raster counter changes at the end of the visible
//display not at the beginning of an actual new raster.

//this register feeds one of the 40 character bytes to the character decoder
reg [319:0] char;

//this register feeds one of the 40 color nybbles to the color decoder
reg [159:0] color;

//holds bitmap data (character or graphics)
reg [319:0] bitmap;
//buffers to save 6*40*25 = 6000 dma cycles per screen display in character display mode
reg [319:0] bitmap2;
reg [319:0] bitmap3;
reg [319:0] bitmap4;
reg [2:0] delayed_bitmap; //for horizontal scrolling

//holds sprite bitmap data for 2nd raster
reg [47:0] spriteData, sprite0Data, sprite1Data, sprite2Data, sprite3Data, sprite4Data, sprite5Data, sprite6Data, sprite7Data;

//sprite bitmap data shift register
reg [47:0] rsprite0Data, rsprite1Data, rsprite2Data, rsprite3Data, rsprite4Data, rsprite5Data, rsprite6Data, rsprite7Data;

//#################################################          
//vic interupt stuff
reg ERST;
reg EMDC;
reg EMMC;
reg ELP;

reg IRST;
reg IMDC;
reg IMMC;
reg ILP;

wire [7:0] irq_register, irq_mask;
wire irq;
assign irq = ~irq_register[7];

assign irq_register[7] = (IRST | IMDC | IMMC | ILP);
assign irq_register[6] = L;
assign irq_register[5] = L;
assign irq_register[4] = L;
assign irq_register[3] = ILP;
assign irq_register[2] = IMMC;
assign irq_register[1] = IMDC;
assign irq_register[0] = IRST;
 
assign irq_mask[7] = L;
assign irq_mask[6] = L;
assign irq_mask[5] = L;
assign irq_mask[4] = L;
assign irq_mask[3] = ELP;
assign irq_mask[2] = EMMC;
assign irq_mask[1] = EMDC;
assign irq_mask[0] = ERST;


//#################################################          
//vic memory registers
reg [8:0] vic_00, vic_02, vic_04, vic_06, vic_08, vic_0A, vic_0C, vic_0E, vic_12;
reg [7:0] vic_01, vic_03, vic_05, vic_07, vic_09, vic_0B, vic_0D, vic_0F;

//screen blanked (border)
//reg [6:0] vic_11 = 7'h0B;
reg [6:0] vic_11 = 7'h1B;

reg [7:0] vic_15;
reg [4:0] vic_16 = 5'h08;
reg [7:0] vic_17;
reg [7:1] vic_18 = 7'b0001010;
reg [7:0] vic_1B, vic_1C, vic_1D, vic_1E, vic_1F, delayed_char;

//border color white
//reg [3:0] vic_20 = 4'h1;
reg [3:0] vic_20 = 4'h3;

reg [3:0] vic_21 = 4'h1;
reg [3:0] vic_22, vic_23, vic_24, vic_25, vic_26, vic_27;
reg [3:0] vic_28, vic_29, vic_2A, vic_2B, vic_2C, vic_2D, vic_2E, delayed_color;

//#################################################          
//vic buffered memory registers
//these registers are updated every 2 raster lines during blanking of active video
//and updated every raster line during blanking of non-active video...
//this is so register values can be changed without affecting the display
reg [8:0] rvic_00, rvic_02, rvic_04, rvic_06, rvic_08, rvic_0A, rvic_0C, rvic_0E;
reg [7:0] rvic_01, rvic_03, rvic_05, rvic_07, rvic_09, rvic_0B, rvic_0D, rvic_0F;//, rvic_10;
reg [6:0] rvic_11;
reg [7:0] rvic_15;
reg [4:0] rvic_16;
reg [7:0] rvic_17;
reg [7:1] rvic_18; 
reg [7:0] rvic_1B, rvic_1C, rvic_1D;//, rvic_1E, rvic_1F;
reg [3:0] rvic_20, rvic_21, rvic_22, rvic_23, rvic_24, rvic_25, rvic_26, rvic_27;
reg [3:0] rvic_28, rvic_29, rvic_2A, rvic_2B, rvic_2C, rvic_2D, rvic_2E;

reg [3:0] videoReg;    //loaded for final processing - convert 4 bit selection to 8 
reg [7:0] videoRegOut; //Drives 8 bit output lines

reg [8:0] shiftHCounter = 0, c64HCounter, c64VCounter;
reg [6:0] dmaCount;
reg [2:0] dmaVCounter = 0;//implements the vscroll capabilities

reg charAccess, charAccess2;

reg r_vDispVisible = L;
reg r_displayHScreen = L;
reg r_displayVScreen = L;


//#################################################          
//memory bus processing

reg [15:0] vic_a;
wire [15:1] vic_a_out = vic_a[15:1];

reg [7:0] charCounter;
reg [9:0] byteCounter;

//do not use the character rom if in bitmap Mode - ram access only
reg charRom;
reg vic_req; 

reg [31:8] spritePointers1; //sprite 0 pointer used straight off memory bus
reg [31:0] spritePointers2;

//these are for aligning the character bitmap accesses
reg [1:0] byteSelect;

//register the color controls
reg r_displayBlack, r_displayBorder, r_displayBG;
reg r_displayBlack2, r_displayBorder2, r_displayBG2;
reg [3:0] r_BG, r_FG;
reg [3:0] r_BG2, r_FG2;
wire [3:0] CharLSN, CharMSN;

wire [7:0] spriteMSB; 

reg shiftSprite0, shiftSprite1, shiftSprite2, shiftSprite3, shiftSprite4, shiftSprite5, shiftSprite6, shiftSprite7;
reg r_displaySP0, r_displaySP1, r_displaySP2, r_displaySP3, r_displaySP4, r_displaySP5, r_displaySP6, r_displaySP7;
reg r_displaySP0_2, r_displaySP1_2, r_displaySP2_2, r_displaySP3_2, r_displaySP4_2, r_displaySP5_2, r_displaySP6_2, r_displaySP7_2;
reg [3:0] r_colorSprite0, r_colorSprite1, r_colorSprite2, r_colorSprite3, r_colorSprite4, r_colorSprite5, r_colorSprite6, r_colorSprite7;
reg [3:0] r_colorSprite0_2, r_colorSprite1_2, r_colorSprite2_2, r_colorSprite3_2, r_colorSprite4_2, r_colorSprite5_2, r_colorSprite6_2, r_colorSprite7_2;

reg [7:0] sprite0YIndex, sprite1YIndex, sprite2YIndex, sprite3YIndex, sprite4YIndex, sprite5YIndex, sprite6YIndex, sprite7YIndex;

reg enableSprite;
reg xExpand;
reg multiColor;

wire [5:0] multiplyBy3;
reg [4:0] multiplyBy3In;
assign multiplyBy3[5:0] = multiplyBy3In[4:0] * 2'h3;

wire MCM, BMM, ECM, nBlank, Row25, Column40;
wire [2:0] ScrollY, ScrollX;

assign ECM = rvic_11[6];
assign BMM = rvic_11[5];
assign nBlank = rvic_11[4];
assign Row25 = rvic_11[3];
assign ScrollY[2:0] = rvic_11[2:0];
assign ScrollX[2:0] = rvic_16[2:0];
assign MCM = rvic_16[4];
assign Column40 = rvic_16[3];

reg dmaEnable = H;

`ifdef mouser
reg [3:0] vic_2F = 4'h2, rvic_2F;//mouseColor;
reg [2:0] mousePixelX, mousePixelY;
reg [0:63] mouse = 64'hFCF8F8DC8E070300;
`endif

//read vic control registers continuously
assign dr[7:0] = dAB[0] ? vic_00[7:0] :
                 dAB[1] ? vic_01[7:0] :
                 dAB[2] ? vic_02[7:0] :
                 dAB[3] ? vic_03[7:0] :
                 dAB[4] ? vic_04[7:0] :
                 dAB[5] ? vic_05[7:0] :
                 dAB[6] ? vic_06[7:0] :
                 dAB[7] ? vic_07[7:0] :
                 dAB[8] ? vic_08[7:0] :
                 dAB[9] ? vic_09[7:0] :
                 dAB[10] ? vic_0A[7:0] :
                 dAB[11] ? vic_0B[7:0] :
                 dAB[12] ? vic_0C[7:0] :
                 dAB[13] ? vic_0D[7:0] :
                 dAB[14] ? vic_0E[7:0] : 
                 dAB[15] ? vic_0F[7:0] :
                 dAB[16] ? spriteMSB :
                 dAB[17] ? {c64VCounter[8],vic_11[6:0]} :
                 dAB[18] ? c64VCounter[7:0] :
                 dAB[19] ? 8'h00 : //vic_13[7:0] :
                 dAB[20] ? 8'h00 : //vic_14[7:0] :
                 dAB[21] ? vic_15[7:0] :
                 dAB[22] ? {3'h0, vic_16[4:0]} :
                 dAB[23] ? vic_17[7:0] :
                 dAB[24] ? {vic_18[7:1], L} :
                 dAB[25] ? irq_register :
                 dAB[26] ? irq_mask :
                 dAB[27] ? vic_1B[7:0] :
                 dAB[28] ? vic_1C[7:0] :
                 dAB[29] ? vic_1D[7:0] :
                 dAB[30] ? vic_1E[7:0] :
                 dAB[31] ? vic_1F[7:0] :
                 dAB[32] ? { 4'h0, vic_20[3:0] } :
                 dAB[33] ? { 4'h0, vic_21[3:0] } :
                 dAB[34] ? { 4'h0, vic_22[3:0] } :
                 dAB[35] ? { 4'h0, vic_23[3:0] } :
                 dAB[36] ? { 4'h0, vic_24[3:0] } :
                 dAB[37] ? { 4'h0, vic_25[3:0] } :
                 dAB[38] ? { 4'h0, vic_26[3:0] } :
                 dAB[39] ? { 4'h0, vic_27[3:0] } :
                 dAB[40] ? { 4'h0, vic_28[3:0] } :
                 dAB[41] ? { 4'h0, vic_29[3:0] } :
                 dAB[42] ? { 4'h0, vic_2A[3:0] } :
                 dAB[43] ? { 4'h0, vic_2B[3:0] } :
                 dAB[44] ? { 4'h0, vic_2C[3:0] } :
                 dAB[45] ? { 4'h0, vic_2D[3:0] } :
                 dAB[46] ? { 4'h0, vic_2E[3:0] } : 
`ifdef mouser
                 dAB[47] ? { 4'h0, vic_2F[3:0] } : 8'hA5;
`else
                 8'hA5;
`endif

reg lp2;

`ifdef la
//logic analyzer signals
wire samplingDone;
wire [4:0] laCounter = vCounter[7:3]; //changes every 8 vga pixels
reg r_displayLA = H;
reg r_displayLA2 = H;
reg r_displayLAx = H;
reg r_displayLAx2 = H;
reg la_sync;
`endif

`ifdef mouser
reg r_displayCursor, r_displayCursor2;
`endif

`ifdef debug
wire dcharout2;
reg dcharout, dcharout1;
reg [2:0] debugHCount;
reg [5:0] debug_Data2;
reg [7:0] debugVCounter;
wire [4:0] debug_Address = debugVCounter[7:3];

//                       char              raster#                   bit#
debug_char dchar(vclk, { debug_Data2[5:0], debugVCounter[2:0], debugHCount[2:0] }, dcharout2, dcharout3);
`endif

//vga timing
wire [8:0] hCounter;
wire [9:0] vCounter;

vgaTiming vga( mclk, vclk, HS, VS, r_displayHBorder, r_displayVBorder, hCounter[8:0], vCounter[9:0] );

always @(posedge mclk)
  begin

    //only update video output when this signal is high
    if(vclk)
      begin    
        
`ifdef debug
        //debug data select
        if(hCounter == 436) debug_Data2 <= debug_Data[47:42];
        else if(hCounter == 444) debug_Data2 <= debug_Data[41:36];
        else if(hCounter == 452) debug_Data2 <= debug_Data[35:30];
        else if(hCounter == 460) debug_Data2 <= debug_Data[29:24];
        else if(hCounter == 468) debug_Data2 <= debug_Data[23:18];
        else if(hCounter == 476) debug_Data2 <= debug_Data[17:12];
        else if(hCounter == 484) debug_Data2 <= debug_Data[11:6];
        else if(hCounter == 492) debug_Data2 <= debug_Data[5:0];
        else if(hCounter == 500) debug_Data2 <= 6'h24; //space

        if(hCounter == 436) debugHCount <= 3'h0;
        else debugHCount <= debugHCount + H;
        
        dcharout <= dcharout1;
        dcharout1 <= debug_sel[0] ? dcharout3 : dcharout2;
`endif

        //end of the scanline
        if(hCounter == 511)
          begin
`ifdef debug
            debugVCounter[7:0] <= vCounter[8:1] + 8'hEF;
`endif
            c64HCounter <= c64HCounter + H;
            
            if(vCounter == 524)
              begin
                c64VCounter <= 9'd12;
              end
            else if(vCounter == 499)
              begin
                c64VCounter <= 9'd0;
              end
            else if(vCounter > 499)
              begin
                c64VCounter <= c64VCounter + vCounter[0];
              end
            else if(~vCounter[0])
              begin
                c64VCounter <= c64VCounter + H;
              end

            if(vCounter == 9'd68)
              begin
                charCounter <= 8'h00;   //character / color ram counter
                byteCounter <= 10'h000; //character or graphic bitmap counter
              end

            // the raster that dma starts on depends on the ScrollY value
            // it continues until the bottom border is reached
            // some of the first rasters may not be displayed depending on the row25 setting
            if(dmaEnable)
              begin
                if((vCounter == 9'd68) & (ScrollY == 3'h0))
                  begin
                    dmaEnable <= L;
                    dmaVCounter <= 3'h7;
                  end
                else if((vCounter == 9'd70) & (ScrollY == 3'h1))
                  begin
                    dmaEnable <= L;
                    dmaVCounter <= 3'h7;
                  end
                else if((vCounter == 9'd72) & (ScrollY == 3'h2))
                  begin
                    dmaEnable <= L;
                    dmaVCounter <= 3'h7;
                  end
                else if((vCounter == 9'd74) & (ScrollY == 3'h3))//default
                  begin
                    dmaEnable <= L;
                    dmaVCounter <= 3'h7;
                  end
                else if((vCounter == 9'd76) & (ScrollY == 3'h4))
                  begin
                    dmaEnable <= L;
                    dmaVCounter <= 3'h7;
                  end
                else if((vCounter == 9'd78) & (ScrollY == 3'h5))
                  begin
                    dmaEnable <= L;
                    dmaVCounter <= 3'h7;
                  end
                else if((vCounter == 9'd80) & (ScrollY == 3'h6))
                  begin
                    dmaEnable <= L;
                    dmaVCounter <= 3'h7;
                  end
                else if((vCounter == 9'd82) & (ScrollY == 3'h7))
                  begin
                    dmaEnable <= L;
                    dmaVCounter <= 3'h7;
                  end
              end

            if(vCounter == 74)
              begin
                if(Row25)
                  begin
                    r_vDispVisible <= H;
                  end
                r_displayVScreen <= H;
              end
            else if(vCounter == 82)
              begin
                r_vDispVisible <= H;
              end
            else if(vCounter == 466)
              begin
                if(~Row25)
                  begin
                    dmaEnable <= H;
                    r_vDispVisible <= L;
                  end
              end
            else if(vCounter == 474)
              begin
                dmaEnable <= H;
                r_vDispVisible <= L;
                r_displayVScreen <= L;
              end
          end
`ifdef debug
        else if(&debug_sel[2:0]) //no debug info - center display
`else
        else
`endif
          begin
            //delay horizontal display based on the ScrollX value
            if(hCounter == (9'd135 + ScrollX[2:0]) & ~dmaEnable)
              begin
                shiftHCounter <= 9'd320;
              end
              
            if(hCounter == 112) c64HCounter <= 0;
            else c64HCounter <= c64HCounter + H;

            if(hCounter == 136)
              begin
                if(Column40)
                  begin
                    r_displayHScreen <= H;
                  end
              end
            else if(hCounter == 143)
              begin
                r_displayHScreen <= H;
              end
            else if(hCounter == 447)
              begin
                if(~Column40)
                  begin
                    r_displayHScreen <= L;
                  end
              end
            else if(hCounter == 456) r_displayHScreen <= L;
          end
`ifdef debug
        else
          begin
            //delay horizontal display based on the ScrollX value
            if(hCounter == (9'd115 + ScrollX[2:0]) & ~dmaEnable) // default ScrollX == 1
              begin
                shiftHCounter <= 9'd320;
              end
              
            if(hCounter == 92) c64HCounter <= 0;
            else c64HCounter <= c64HCounter + H;

            if(hCounter == 116)
              begin
                if(Column40)
                  begin
                    r_displayHScreen <= H;
                  end
              end
            else if(hCounter == 123)
              begin
                r_displayHScreen <= H;
              end
            else if(hCounter == 427)
              begin
                if(~Column40)
                  begin
                    r_displayHScreen <= L;
                  end
              end
            else if(hCounter == 436) r_displayHScreen <= L;
          end
`endif
            
`ifdef la
       //logic analyzer data stream processing
        if(c64HCounter == 510) la_sync <= L;
        else if(c64HCounter == 511) la_sync <= H;

        if(~samplingDone & (vCounter > 259))
          begin
            if(vCounter[2:0] == 4)
              begin
                r_displayLA <= lain;
                r_displayLAx <= ~lain;
              end
            else if(vCounter[2:0] == 7)
              begin
                r_displayLA <= ~lain;
                r_displayLAx <= lain;
              end
            else
              begin
                r_displayLA <= H;
                r_displayLAx <= H;
              end
          end
        else
          begin
            r_displayLA <= H;
            r_displayLAx <= H;
          end

        r_displayLA2 <= r_displayLA;
        r_displayLAx2 <= r_displayLAx;
`endif
        
`ifdef mouser
        r_displayCursor <= ((c64HCounter >= mouse_x) & (c64HCounter < (mouse_x + 9'h008)) & (c64VCounter >= { L, mouse_y }) & (c64VCounter < ({ L, mouse_y } + 9'h008)));
        mousePixelY <= {c64VCounter - mouse_y}[2:0];
        mousePixelX <= {c64HCounter - mouse_x}[2:0];
        
        r_displayCursor2 <= r_displayCursor & mouse[{ mousePixelY, mousePixelX }];
`endif

        //all display values must be registered the same amount...
        
        //force displaying black when not in an area of active video
        r_displayBlack <= ~(r_displayVBorder && r_displayHBorder);
        
        //force displaying the border color if in the border or the screen is blanked
        r_displayBorder <= ~(r_vDispVisible && r_displayHScreen) | ~nBlank;
    
        //select displaying the background color over the foreground color
        r_displayBG <= ~(delayed_bitmap[2] | ECM & ((~hCounter[0] & delayed_bitmap[1]) | (hCounter[0] & delayed_bitmap[0])));
    
        //selected foreground color
        r_FG <= BMM & ~MCM ? CharMSN :
                BMM &  MCM & delayed_bitmap[2] & ~delayed_bitmap[1] ? CharLSN : delayed_color;//color[159:156];

        //select background color
        r_BG <= ~BMM &  ECM & char[319] &  char[318] ? rvic_24 : 
                ~BMM &  ECM & char[319] & ~char[318] ? rvic_23 : 
                ~BMM &  ECM & ~char[319] & char[318] ? rvic_22 :
                ~BMM &  ECM & ~char[319] & ~char[318] ? rvic_21 :
                 BMM & ~MCM ? CharLSN :
                 BMM &  MCM & ~delayed_bitmap[2] & delayed_bitmap[1] ? CharMSN : rvic_21;
    
        //sprite showing (non-transparent pixel)
        r_displaySP0 <= shiftSprite0 & (rsprite0Data[47] | (rvic_1C[0] & rsprite0Data[46]));
        r_displaySP1 <= shiftSprite1 & (rsprite1Data[47] | (rvic_1C[1] & rsprite1Data[46]));
        r_displaySP2 <= shiftSprite2 & (rsprite2Data[47] | (rvic_1C[2] & rsprite2Data[46]));
        r_displaySP3 <= shiftSprite3 & (rsprite3Data[47] | (rvic_1C[3] & rsprite3Data[46]));
        r_displaySP4 <= shiftSprite4 & (rsprite4Data[47] | (rvic_1C[4] & rsprite4Data[46]));
        r_displaySP5 <= shiftSprite5 & (rsprite5Data[47] | (rvic_1C[5] & rsprite5Data[46]));
        r_displaySP6 <= shiftSprite6 & (rsprite6Data[47] | (rvic_1C[6] & rsprite6Data[46]));
        r_displaySP7 <= shiftSprite7 & (rsprite7Data[47] | (rvic_1C[7] & rsprite7Data[46]));

        //sprite color
        r_colorSprite0 <= (rvic_1C[0] & rsprite0Data[47] & rsprite0Data[46]) ? rvic_26[3:0] : (rvic_1C[0] & ~rsprite0Data[47] & rsprite0Data[46]) ? rvic_25[3:0] : rvic_27[3:0]; 
        r_colorSprite1 <= (rvic_1C[1] & rsprite1Data[47] & rsprite1Data[46]) ? rvic_26[3:0] : (rvic_1C[1] & ~rsprite1Data[47] & rsprite1Data[46]) ? rvic_25[3:0] : rvic_28[3:0]; 
        r_colorSprite2 <= (rvic_1C[2] & rsprite2Data[47] & rsprite2Data[46]) ? rvic_26[3:0] : (rvic_1C[2] & ~rsprite2Data[47] & rsprite2Data[46]) ? rvic_25[3:0] : rvic_29[3:0]; 
        r_colorSprite3 <= (rvic_1C[3] & rsprite3Data[47] & rsprite3Data[46]) ? rvic_26[3:0] : (rvic_1C[3] & ~rsprite3Data[47] & rsprite3Data[46]) ? rvic_25[3:0] : rvic_2A[3:0]; 
        r_colorSprite4 <= (rvic_1C[4] & rsprite4Data[47] & rsprite4Data[46]) ? rvic_26[3:0] : (rvic_1C[4] & ~rsprite4Data[47] & rsprite4Data[46]) ? rvic_25[3:0] : rvic_2B[3:0]; 
        r_colorSprite5 <= (rvic_1C[5] & rsprite5Data[47] & rsprite5Data[46]) ? rvic_26[3:0] : (rvic_1C[5] & ~rsprite5Data[47] & rsprite5Data[46]) ? rvic_25[3:0] : rvic_2C[3:0]; 
        r_colorSprite6 <= (rvic_1C[6] & rsprite6Data[47] & rsprite6Data[46]) ? rvic_26[3:0] : (rvic_1C[6] & ~rsprite6Data[47] & rsprite6Data[46]) ? rvic_25[3:0] : rvic_2D[3:0]; 
        r_colorSprite7 <= (rvic_1C[7] & rsprite7Data[47] & rsprite7Data[46]) ? rvic_26[3:0] : (rvic_1C[7] & ~rsprite7Data[47] & rsprite7Data[46]) ? rvic_25[3:0] : rvic_2E[3:0]; 

        //delay all display values so more selections can be made...
        r_displayBlack2 <= r_displayBlack;
        r_displayBorder2 <= r_displayBorder;
        
        //not done...
        r_displayBG2 <= r_displayBG;
    
        //not done...
        // the selected foreground color must be registered
        r_FG2 <= r_FG;

        //not done...
        r_BG2 <= r_BG;

        //sprite showing (non-transparent pixel)
        r_displaySP0_2 <= r_displaySP0;
        r_displaySP1_2 <= r_displaySP1;
        r_displaySP2_2 <= r_displaySP2;
        r_displaySP3_2 <= r_displaySP3;
        r_displaySP4_2 <= r_displaySP4;
        r_displaySP5_2 <= r_displaySP5;
        r_displaySP6_2 <= r_displaySP6;
        r_displaySP7_2 <= r_displaySP7;

        //sprite color
        r_colorSprite0_2 <= r_displaySP0_2 & ~rvic_1B[0] & ~r_displayBG2 ? r_FG[3:0] : r_colorSprite0[3:0]; 
        r_colorSprite1_2 <= r_displaySP1_2 & ~rvic_1B[1] & ~r_displayBG2 ? r_FG[3:0] : r_colorSprite1[3:0]; 
        r_colorSprite2_2 <= r_displaySP2_2 & ~rvic_1B[2] & ~r_displayBG2 ? r_FG[3:0] : r_colorSprite2[3:0]; 
        r_colorSprite3_2 <= r_displaySP3_2 & ~rvic_1B[3] & ~r_displayBG2 ? r_FG[3:0] : r_colorSprite3[3:0]; 
        r_colorSprite4_2 <= r_displaySP4_2 & ~rvic_1B[4] & ~r_displayBG2 ? r_FG[3:0] : r_colorSprite4[3:0]; 
        r_colorSprite5_2 <= r_displaySP5_2 & ~rvic_1B[5] & ~r_displayBG2 ? r_FG[3:0] : r_colorSprite5[3:0]; 
        r_colorSprite6_2 <= r_displaySP6_2 & ~rvic_1B[6] & ~r_displayBG2 ? r_FG[3:0] : r_colorSprite6[3:0]; 
        r_colorSprite7_2 <= r_displaySP7_2 & ~rvic_1B[7] & ~r_displayBG2 ? r_FG[3:0] : r_colorSprite7[3:0]; 

        //register color output to avoid logic glitches
        videoReg <= r_displayBlack2 ? 4'h0 : 
`ifdef mouser
                    r_displayCursor2 ? rvic_2F : 
`endif
`ifdef la
                    //the first valid la data shows up 7 cycles after la_sync
                    (~r_displayLA2 & (c64HCounter > 0) & (c64HCounter < 391)) ? 4'h0 : 
                    (~r_displayLAx2 & (c64HCounter > 0) & (c64HCounter < 391)) ? 4'hF : 
`endif
`ifdef debug
                    ((hCounter > 438) & (hCounter < 511) & dcharout) ? rvic_20[3:0] ^ 4'hF :
`endif
                    r_displayBorder2 ? rvic_20[3:0] :
                    r_displaySP0_2 ? r_colorSprite0_2 :
                    r_displaySP1_2 ? r_colorSprite1_2 :
                    r_displaySP2_2 ? r_colorSprite2_2 :
                    r_displaySP3_2 ? r_colorSprite3_2 :
                    r_displaySP4_2 ? r_colorSprite4_2 :
                    r_displaySP5_2 ? r_colorSprite5_2 :
                    r_displaySP6_2 ? r_colorSprite6_2 :
                    r_displaySP7_2 ? r_colorSprite7_2 :
                    r_displayBG2 ? r_BG2[3:0] : r_FG2[3:0];

        //sprite Y index
        sprite0YIndex <= vCounter[8:1] - rvic_01[7:0];
        sprite1YIndex <= vCounter[8:1] - rvic_03[7:0];
        sprite2YIndex <= vCounter[8:1] - rvic_05[7:0];
        sprite3YIndex <= vCounter[8:1] - rvic_07[7:0];
        sprite4YIndex <= vCounter[8:1] - rvic_09[7:0];
        sprite5YIndex <= vCounter[8:1] - rvic_0B[7:0];
        sprite6YIndex <= vCounter[8:1] - rvic_0D[7:0];
        sprite7YIndex <= vCounter[8:1] - rvic_0F[7:0];

        //color lookup table
        case(videoReg)
          4'h0: videoRegOut <= color_black;
          4'h1: videoRegOut <= color_white;
          4'h2: videoRegOut <= color_red;
          4'h3: videoRegOut <= color_cyan;
          4'h4: videoRegOut <= color_purple;
          4'h5: videoRegOut <= color_green;
          4'h6: videoRegOut <= color_blue;
          4'h7: videoRegOut <= color_yellow;
          4'h8: videoRegOut <= color_orange;
          4'h9: videoRegOut <= color_brown;
          4'hA: videoRegOut <= color_Lred;
          4'hB: videoRegOut <= color_Dgray;
          4'hC: videoRegOut <= color_Mgray;
          4'hD: videoRegOut <= color_Lgreen;
          4'hE: videoRegOut <= color_Lblue;
          4'hF: videoRegOut <= color_Lgray;
        endcase

      end

    //this stuff is happennig every memory cycle
    lp2 <= LP;

    if(~we)
      begin
        //write to vic registers
        if(dAB[0])  vic_00[7:0] <= dw[7:0]; //Sprite 0 X Pos
        if(dAB[1])  vic_01[7:0] <= dw[7:0]; //Sprite 0 Y Pos
        if(dAB[2])  vic_02[7:0] <= dw[7:0]; //Sprite 1 X Pos
        if(dAB[3])  vic_03[7:0] <= dw[7:0]; //Sprite 1 Y Pos
        if(dAB[4])  vic_04[7:0] <= dw[7:0]; //Sprite 2 X Pos
        if(dAB[5])  vic_05[7:0] <= dw[7:0]; //Sprite 2 Y Pos
        if(dAB[6])  vic_06[7:0] <= dw[7:0]; //Sprite 3 X Pos
        if(dAB[7])  vic_07[7:0] <= dw[7:0]; //Sprite 3 Y Pos
        if(dAB[8])  vic_08[7:0] <= dw[7:0]; //Sprite 4 X Pos
        if(dAB[9])  vic_09[7:0] <= dw[7:0]; //Sprite 4 Y Pos
        if(dAB[10]) vic_0A[7:0] <= dw[7:0]; //Sprite 5 X Pos
        if(dAB[11]) vic_0B[7:0] <= dw[7:0]; //Sprite 5 Y Pos
        if(dAB[12]) vic_0C[7:0] <= dw[7:0]; //Sprite 6 X Pos
        if(dAB[13]) vic_0D[7:0] <= dw[7:0]; //Sprite 6 Y Pos
        if(dAB[14]) vic_0E[7:0] <= dw[7:0]; //Sprite 7 X Pos
        if(dAB[15]) vic_0F[7:0] <= dw[7:0]; //Sprite 7 Y Pos

        if(dAB[16])
          begin
            vic_00[8] <= dw[0]; //Sprite 0 X Pos msb
            vic_02[8] <= dw[1]; //Sprite 1 X Pos msb
            vic_04[8] <= dw[2]; //Sprite 2 X Pos msb
            vic_06[8] <= dw[3]; //Sprite 3 X Pos msb
            vic_08[8] <= dw[4]; //Sprite 4 X Pos msb
            vic_0A[8] <= dw[5]; //Sprite 5 X Pos msb
            vic_0C[8] <= dw[6]; //Sprite 6 X Pos msb
            vic_0E[8] <= dw[7]; //Sprite 7 X Pos msb
          end

        if(dAB[17])
          begin
            vic_12[8] <= dw[7]; //Raster Register msb
            vic_11[6:0] <= dw[6:0];
          end

        if(dAB[18]) vic_12[7:0] <= dw[7:0]; //Raster Register

        if(dAB[21]) vic_15[7:0] <= dw[7:0];
        if(dAB[22]) vic_16[4:0] <= dw[4:0];
        if(dAB[23]) vic_17[7:0] <= dw[7:0];
        if(dAB[24]) vic_18[7:1] <= dw[7:1];

//currently unused
//        if(dAB[25]) vic_19[7:1] <= dw[7:1];
//        if(dAB[26]) vic_1A[7:1] <= dw[7:1];

        if(dAB[27]) vic_1B[7:0] <= dw[7:0];
        if(dAB[28]) vic_1C[7:0] <= dw[7:0];
        if(dAB[29]) vic_1D[7:0] <= dw[7:0];

        if(dAB[32]) vic_20[3:0] <= dw[3:0];
        if(dAB[33]) vic_21[3:0] <= dw[3:0];
        if(dAB[34]) vic_22[3:0] <= dw[3:0];
        if(dAB[35]) vic_23[3:0] <= dw[3:0];
        if(dAB[36]) vic_24[3:0] <= dw[3:0];
        if(dAB[37]) vic_25[3:0] <= dw[3:0];
        if(dAB[38]) vic_26[3:0] <= dw[3:0];
        if(dAB[39]) vic_27[3:0] <= dw[3:0];
        if(dAB[40]) vic_28[3:0] <= dw[3:0];
        if(dAB[41]) vic_29[3:0] <= dw[3:0];
        if(dAB[42]) vic_2A[3:0] <= dw[3:0];
        if(dAB[43]) vic_2B[3:0] <= dw[3:0];
        if(dAB[44]) vic_2C[3:0] <= dw[3:0];
        if(dAB[45]) vic_2D[3:0] <= dw[3:0];
        if(dAB[46]) vic_2E[3:0] <= dw[3:0];
`ifdef mouser
        if(dAB[47])
          begin
            vic_2F[3:0] <= dw[3:0];
            mouse <= mouse ^ 64'hFFFFFFFFFFFFFFFF;
          end
`endif
      end

    if(~we & dAB[25])
      begin
        IRST <= dw[0] ? L : ERST & (vCounter[9:1] == vic_12[8:0]);
        IMDC <= dw[1] ? L : EMDC & (vic_1E[0] | vic_1E[1] | vic_1E[2] | vic_1E[3] | vic_1E[4] | vic_1E[5] | vic_1E[6] | vic_1E[7]);
        IMMC <= dw[2] ? L : EMMC & (vic_1F[0] | vic_1F[1] | vic_1F[2] | vic_1F[3] | vic_1F[4] | vic_1F[5] | vic_1F[6] | vic_1F[7]);
        ILP <= dw[3] ? L : ELP & ~LP & lp2;
      end
    else if(~reset)
      begin
        IRST <= L;
        IMDC <= L;
        IMMC <= L;
        ILP <= L;
      end
    else
      begin
        IRST <= ERST & (vCounter[9:1] == vic_12[8:0]);
        IMDC <= EMDC & (vic_1E[0] | vic_1E[1] | vic_1E[2] | vic_1E[3] | vic_1E[4] | vic_1E[5] | vic_1E[6] | vic_1E[7]);
        IMMC <= EMMC & (vic_1F[0] | vic_1F[1] | vic_1F[2] | vic_1F[3] | vic_1F[4] | vic_1F[5] | vic_1F[6] | vic_1F[7]);
        ILP <= ELP & ~LP & lp2;
      end

    if(~we & dAB[26])
      begin
        ERST <= dw[0]; 
        EMDC <= dw[1];
        EMMC <= dw[2];
        ELP <= dw[3];
      end
    else if(~reset)
      begin
        ERST <= L; 
        EMDC <= L;
        EMMC <= L;
        ELP <= L;
      end

    //sprite to sprite collisions - this may not be right
    vic_1E[0] <= ~re & dAB[30] ? L : vic_1E[0] | r_displaySP0 & (r_displaySP1 | r_displaySP2 | r_displaySP3 | shiftSprite4 | shiftSprite5 | shiftSprite6 | shiftSprite7);
    vic_1E[1] <= ~re & dAB[30] ? L : vic_1E[1] | r_displaySP1 & (r_displaySP0 | r_displaySP2 | r_displaySP3 | shiftSprite4 | shiftSprite5 | shiftSprite6 | shiftSprite7);
    vic_1E[2] <= ~re & dAB[30] ? L : vic_1E[2] | r_displaySP2 & (r_displaySP0 | r_displaySP1 | r_displaySP3 | shiftSprite4 | shiftSprite5 | shiftSprite6 | shiftSprite7);
    vic_1E[3] <= ~re & dAB[30] ? L : vic_1E[3] | r_displaySP3 & (r_displaySP0 | r_displaySP1 | r_displaySP2 | shiftSprite4 | shiftSprite5 | shiftSprite6 | shiftSprite7);
    vic_1E[4] <= ~re & dAB[30] ? L : vic_1E[4] | r_displaySP4 & (r_displaySP0 | r_displaySP1 | r_displaySP2 | r_displaySP3 | shiftSprite5 | shiftSprite6 | shiftSprite7);
    vic_1E[5] <= ~re & dAB[30] ? L : vic_1E[5] | r_displaySP5 & (r_displaySP0 | r_displaySP1 | r_displaySP2 | r_displaySP3 | shiftSprite4 | shiftSprite6 | shiftSprite7);
    vic_1E[6] <= ~re & dAB[30] ? L : vic_1E[6] | r_displaySP6 & (r_displaySP0 | r_displaySP1 | r_displaySP2 | r_displaySP3 | shiftSprite4 | shiftSprite5 | shiftSprite7);
    vic_1E[7] <= ~re & dAB[30] ? L : vic_1E[7] | r_displaySP7 & (r_displaySP0 | r_displaySP1 | r_displaySP2 | r_displaySP3 | shiftSprite4 | shiftSprite5 | shiftSprite6);
    
    //sprite to foreground collisions - this is not right
    vic_1F[0] <= ~re & dAB[31] ? L : vic_1E[0] | r_displaySP0;// & r_FG[0];
    vic_1F[1] <= ~re & dAB[31] ? L : vic_1E[1] | r_displaySP1;// & r_FG[1];
    vic_1F[2] <= ~re & dAB[31] ? L : vic_1E[2] | r_displaySP2;// & r_FG[2];
    vic_1F[3] <= ~re & dAB[31] ? L : vic_1E[3] | r_displaySP3;// & r_FG[3];
    vic_1F[4] <= ~re & dAB[31] ? L : vic_1E[4] | r_displaySP4;// & r_FG[4];
    vic_1F[5] <= ~re & dAB[31] ? L : vic_1E[5] | r_displaySP5;// & r_FG[5];
    vic_1F[6] <= ~re & dAB[31] ? L : vic_1E[6] | r_displaySP6;// & r_FG[6];
    vic_1F[7] <= ~re & dAB[31] ? L : vic_1E[7] | r_displaySP7;// & r_FG[7];
    

    //nothing can change on the screen during dma cycles
    if(dmaCount != 7'h00)
      begin
        //save lower bits of address so pipelined data can be pulled off
        //of the memory bus correctly...
        byteSelect[1:0] <= vic_a[1:0];
        //dma state machine counter
        dmaCount <= dmaCount - H;
        
//*********************************
//generate addresses
        case(dmaCount)
//loading character pointers and color ram
          65:
            begin
              //determines which raster is the first and middle of each character row
              charAccess <= ~dmaVCounter[2] & ~dmaVCounter[1] & ~dmaVCounter[0];
              charAccess2 <= dmaVCounter[2] & ~dmaVCounter[1] & ~dmaVCounter[0];
            end

          64:
            begin
              //if not necessary this will not start a dma sequence
              vic_req <= ~(r_displayVScreen & charAccess & nBlank);

//the first 10 states only run at the starting raster of a new character...
//this first access is ignored, delay a clock for vic_ack...
                
              vic_a[14] <= ~va14; 
              vic_a[15] <= ~va15; 
              vic_a[13:10] <= rvic_18[7:4];
              vic_a[9:2] <= charCounter[7:0];
              vic_a[1:0] <= 2'h0;
              charRom <= H; //character pointers in memory
            end

          63,62,61,60,59,58,57,56,55:
            begin
              vic_a[9:2] <= charCounter[7:0];
              if(r_displayVScreen & charAccess & nBlank)
                begin
                  charCounter <= charCounter + H;
                end
            end
        
          54:
            begin
              //turn dma on if needed
              vic_req <= ~(r_displayVScreen & (charAccess | charAccess2 | BMM) & nBlank);
              vic_a[9:2] <= charCounter[7:0];
              if(r_displayVScreen & charAccess & nBlank)
                begin
                  charCounter <= charCounter + H;
                end
            end
        
// load the 40 character data or bitmap bytes required ***this could be cut down to 20 using port B with a port A+8 address
// *** or change to save all 8 consecutive bytes for the next 8 (16) rasters? neither of these may not be vic compatible
// because the character rom / bitmap source can be changed on a per raster basis or in the middle of a raster...
          53:
            begin
              if(BMM)
                begin
                  //loading graphics mode bitmap values
                  byteCounter <= byteCounter + H;
                  vic_a[13] <= rvic_18[3];
                  vic_a[12:3] <= byteCounter[9:0];
                  vic_a[2:0] <= dmaVCounter[2:0];
                 end
              else
                begin
                  //loading character bitmap values
                  vic_a[14] <= (va14 & ~rvic_18[3] & rvic_18[2]) | ~va14; 
                  vic_a[15] <= (va14 & ~rvic_18[3] & rvic_18[2]) | ~va15; 

                  //force character rom when necessary
                  //this must only go low when forcing the character rom...
                  charRom <= ~(va14 & ~rvic_18[3] & rvic_18[2]);
                  
                  vic_a[13:11] <= rvic_18[3:1];
                  //raster # per character row
                  vic_a[2:0] <= dmaVCounter[2:0];
 
                  //loading 1st character bitmap values
                  if(r_displayVScreen & charAccess & nBlank)
                    begin
                      //not all character pointers have been loaded yet so these indexes are adjusted
                      //raster 1 & 4 of the 8 character rasters
                      vic_a[10] <= (char[255] & ~MCM);
                      vic_a[9] <= (char[254] & ~MCM);
                      vic_a[8:3] <= char[253:248];
                    end
                  else if(r_displayVScreen & nBlank)
                    begin
                      //extended color text forces only 64 possible character images
                      vic_a[10] <= (char[319] & ~MCM);
                      vic_a[9] <= (char[318] & ~MCM);
                      vic_a[8:3] <= char[317:312];

                      //rotate through the 40 character pointers - needed for multi-color character mode only
                      char <= char << 8;
                      char[7:0] <= char[319:312];
                    end
                end
            end

          52:
            begin
              if(BMM)
                begin
                  //loading graphics mode bitmap values
                  byteCounter <= byteCounter + H;
                  vic_a[12:3] <= byteCounter[9:0];
                end
              else
                begin
                  //loading 2nd character bitmap values
                  if(r_displayVScreen & charAccess & nBlank)
                    begin
                      //not all character pointers have been loaded yet so these indexes are adjusted
                      //for this and the next 2 states...
                      //raster 1 & 4 of the 8 character rasters
                      vic_a[10] <= (char[279] & ~MCM);
                      vic_a[9] <= (char[278] & ~MCM);
                      vic_a[8:3] <= char[277:272];
                    end
                  else if(r_displayVScreen & nBlank)
                    begin
                      //loading character bitmap values
                      //extended color text forces only 64 possible character images
                      vic_a[10] <= (char[319] & ~MCM);
                      vic_a[9] <= (char[318] & ~MCM);
                      vic_a[8:3] <= char[317:312];

                      //rotate through the 40 character pointers - needed for multi-color character mode only
                      char <= char << 8;
                      char[7:0] <= char[319:312];
                    end 
                end
            end
        
          51:
            begin
              if(BMM)
                begin
                  //loading graphics mode bitmap values
                  byteCounter <= byteCounter + H;
                  vic_a[12:3] <= byteCounter[9:0];
                end
              else
                begin
                  //loading 3rd character bitmap values
                  if(r_displayVScreen & charAccess & nBlank)
                    begin
                      //not all character pointers have been loaded yet so these indexes are adjusted
                      //raster 1 & 4 of the 8 character rasters
                      vic_a[10] <= (char[303] & ~MCM);
                      vic_a[9] <= (char[302] & ~MCM);
                      vic_a[8:3] <= char[301:296];

                      //rotate through the 40 character pointers - needed for multi-color character mode only
                      //catch up missed shifts due to pointer loading...
                      char <= char << 24;
                      char[23:0] <= char[319:296];
                    end
                  else if(r_displayVScreen & nBlank)
                    begin
                      //extended color text forces only 64 possible character images
                      vic_a[10] <= (char[319] & ~MCM);
                      vic_a[9] <= (char[318] & ~MCM);
                      vic_a[8:3] <= char[317:312];

                      //rotate through the 40 character pointers - needed for multi-color character mode only
                      char <= char << 8;
                      char[7:0] <= char[319:312];
                    end
                 end
            end

//load character or bitmap data
          default:  //50-15
            begin
              if(BMM)
                begin
                  //loading graphics mode bitmap values
                  byteCounter <= byteCounter + H;
                  vic_a[12:3] <= byteCounter[9:0];
                end
              else
                begin
                  //loading character bitmap values
                  //extended color text forces only 64 possible character images
                  vic_a[10] <= (char[319] & ~MCM);
                  vic_a[9] <= (char[318] & ~MCM);
                  vic_a[8:3] <= char[317:312];

                  //rotate through the 40 character pointers - needed for multi-color character mode only
                  char <= char << 8;
                  char[7:0] <= char[319:312];
                 end
            end

         14:
            begin
              vic_req <= ~(r_displayVScreen & nBlank);

              if(BMM)
                begin
                  //loading graphics mode bitmap values
                  byteCounter <= byteCounter + H;
                  vic_a[12:3] <= byteCounter[9:0];
                end
              else
                begin
                  //loading character bitmap values
                  //extended color text forces only 64 possible character images
                  vic_a[10] <= (char[319] & ~MCM);
                  vic_a[9] <= (char[318] & ~MCM);
                  vic_a[8:3] <= char[317:312];

                  //rotate through the 40 character pointers - needed for multi-color character mode only
                  char <= char << 8;
                  char[7:0] <= char[319:312];
                 end
            end

//loading sprite pointers
          13:
            begin
              charRom <= H; //no more character rom accesses
              //sprite pointers 1-4 address
              vic_a[9:2] <= 8'hFE;
            end

          12:
            begin
              //sprite pointers 5-8 address
              vic_a[9:2] <= 8'hFF;
              multiplyBy3In[4:0] <= (rvic_17[0] ? sprite0YIndex[5:1] : sprite0YIndex[4:0]);
            end

//generate sprite data addresses
          11:
            begin
              //sprite 0 data address
//              vic_a[13:6] <= spritePointers1[7:0]; - sprite pointer1 reg not loaded yet...
              vic_a[13:0] <= { memory[7:0], multiplyBy3[5:0] };
              multiplyBy3In[4:0] <= (rvic_17[1] ? sprite1YIndex[5:1] : sprite1YIndex[4:0]);
            end

          10:
            begin
              //sprite 1 data address
              vic_a[13:0] <= { spritePointers1[15:8], multiplyBy3[5:0] };
              multiplyBy3In[4:0] <= (rvic_17[2] ? sprite2YIndex[5:1] : sprite2YIndex[4:0]);
            end
        
          9:
            begin
              //sprite 2 data address
               vic_a[13:0] <= { spritePointers1[23:16], multiplyBy3[5:0] };
              multiplyBy3In[4:0] <= (rvic_17[3] ? sprite3YIndex[5:1] : sprite3YIndex[4:0]);
            end
        
          8:
            begin
              //sprite 3 data address
              vic_a[13:0] <= { spritePointers1[31:24], multiplyBy3[5:0] };
              multiplyBy3In[4:0] <= (rvic_17[4] ? sprite4YIndex[5:1] : sprite4YIndex[4:0]);
            end
           
          7:
            begin
              //sprite 4 data address
              vic_a[13:0] <= { spritePointers2[7:0], multiplyBy3[5:0] };
              multiplyBy3In[4:0] <= (rvic_17[5] ? sprite5YIndex[5:1] : sprite5YIndex[4:0]);
            end
        
          6:
            begin
              //sprite 5 data address
              vic_a[13:0] <= { spritePointers2[15:8], multiplyBy3[5:0] };
              multiplyBy3In[4:0] <= (rvic_17[6] ? sprite6YIndex[5:1] : sprite6YIndex[4:0]);
            end
        
          5:
            begin
              //sprite 6 data address
              vic_a[13:0] <= { spritePointers2[23:16], multiplyBy3[5:0] };
              multiplyBy3In[4:0] <= (rvic_17[7] ? sprite7YIndex[5:1] : sprite7YIndex[4:0]);
            end
        
          4:
            begin
              //sprite 7 data address
              vic_a[13:0] <= { spritePointers2[31:24], multiplyBy3[5:0] };
            end
        
          3:
            begin
              vic_req <= H; // stop dma cycles
            end

          2,1,0:
            begin
            end
          
        endcase

//*********************************
//load data
        case(dmaCount)

          61,60,59,58,57,56,55,54,53,52:
            begin
              if(r_displayVScreen & charAccess & nBlank)
                begin
                  char <= char << 32;
                  color <= color << 16;
                  char[31:0] <= { memory[7:0], memory[15:8], memory[23:16], memory[31:24] };
                  color[15:0] <= { vic_color[3:0], vic_color[7:4], vic_color[11:8], vic_color[15:12] };
                end
            end

//load character or bitmap data
          default:  //51-12
            begin
              if(BMM)
                begin
                  bitmap <= bitmap << 8;
                  case(byteSelect[1:0])
                    0: bitmap[7:0] <= memory[7:0];
                    1: bitmap[7:0] <= memory[15:8];
                    2: bitmap[7:0] <= memory[23:16];
                    3: bitmap[7:0] <= memory[31:24];
                  endcase
                end
              else if(charAccess | charAccess2)
                begin
                  bitmap <= bitmap << 8;
                  bitmap[7:0] <= memory[7:0];
                  bitmap2 <= bitmap2 << 8;
                  bitmap2[7:0] <= memory[15:8];
                  bitmap3 <= bitmap3 << 8;
                  bitmap3[7:0] <= memory[23:16];
                  bitmap4 <= bitmap4 << 8;
                  bitmap4[7:0] <= memory[31:24];
                end
            end

          11:
            begin
              if(~BMM & ~charAccess & ~charAccess2)
                begin
                  if(~dmaVCounter[1] & dmaVCounter[0])
                    begin
                      bitmap <= bitmap2;
                    end
                  else if(dmaVCounter[1] & ~dmaVCounter[0])
                    begin
                      bitmap <= bitmap3;
                    end
                  else if(dmaVCounter[1] & dmaVCounter[0])
                    begin
                      bitmap <= bitmap4;
                    end
                end

              //sprite pointers 3-1 load
              spritePointers1[31:8] <= memory[31:8];
            end
        
          10:
            begin
              //sprite pointers 7-4 load
              spritePointers2 <= memory[31:0];
            end
        
          9:
            begin

//*****DATA - sprite data load is delayed 1 cycle for  x-expansion...

//preparing for sprite 0 incoming data in the next state
              enableSprite <= rvic_15[0];
              xExpand <= rvic_1D[0];
              multiColor <= rvic_1C[0];
            end
           
          8:
            begin
              //sprite 0 data load
              sprite0Data <= spriteData;
              
//preparing for sprite 1 incoming data in the next state
              enableSprite <= rvic_15[1];
              xExpand <= rvic_1D[1];
              multiColor <= rvic_1C[1];
            end
        
          7:
            begin
              //sprite 1 data load
              sprite1Data <= spriteData;

//preparing for sprite 2 incoming data in the next state
              enableSprite <= rvic_15[2];
              xExpand <= rvic_1D[2];
              multiColor <= rvic_1C[2];
            end
        
          6:
            begin
              //sprite 2 data load
              sprite2Data <= spriteData;

//preparing for sprite 3 incoming data in the next state
              enableSprite <= rvic_15[3];
              xExpand <= rvic_1D[3];
              multiColor <= rvic_1C[3];
            end
        
          5:
            begin
              //sprite 3 data load
              sprite3Data <= spriteData;

//preparing for sprite 4 incoming data in the next state
              enableSprite <= rvic_15[4];
              xExpand <= rvic_1D[4];
              multiColor <= rvic_1C[4];
            end
        
          4:
            begin
              //sprite 4 data load
              sprite4Data <= spriteData;

//preparing for sprite 5 incoming data in the next state
              enableSprite <= rvic_15[5];
              xExpand <= rvic_1D[5];
              multiColor <= rvic_1C[5];
            end

          3:
            begin
              //sprite 5 data load
              sprite5Data <= spriteData;

//preparing for sprite 6 incoming data in the next state
              enableSprite <= rvic_15[6];
              xExpand <= rvic_1D[6];
              multiColor <= rvic_1C[6];
            end
        
          2:
            begin
              //sprite 6 data load
              sprite6Data <= spriteData;

//preparing for sprite 7 incoming data in the next state
              enableSprite <= rvic_15[7];
              xExpand <= rvic_1D[7];
              multiColor <= rvic_1C[7];
            end

          1:
            begin
              //sprite 7 data load
              sprite7Data <= spriteData;

              //sprite X data load enable
              enableSprite <= L;
            end

          65,64,63,62,0:
            begin
            end

        endcase

        //this happens all the time as a generic load operation
        if(~enableSprite)
          begin
            //just make it all transparent
            spriteData[47:0] <= 48'h000000000000;
          end
        else if(xExpand)
          begin
            //X expanded sprite
            if(multiColor)
              begin
                case(byteSelect[1:0])
                  0:
                    begin
                      spriteData[47] <= memory[7];
                      spriteData[46] <= memory[6];
                      spriteData[45] <= memory[7];
                      spriteData[44] <= memory[6];
                      spriteData[43] <= memory[5];
                      spriteData[42] <= memory[4];
                      spriteData[41] <= memory[5];
                      spriteData[40] <= memory[4];
                      spriteData[39] <= memory[3];
                      spriteData[38] <= memory[2];
                      spriteData[37] <= memory[3];
                      spriteData[36] <= memory[2];
                      spriteData[35] <= memory[1];
                      spriteData[34] <= memory[0];
                      spriteData[33] <= memory[1];
                      spriteData[32] <= memory[0];
                      
                      spriteData[31] <= memory[15];
                      spriteData[30] <= memory[14];
                      spriteData[29] <= memory[15];
                      spriteData[28] <= memory[14];
                      spriteData[27] <= memory[13];
                      spriteData[26] <= memory[12];
                      spriteData[25] <= memory[13];
                      spriteData[24] <= memory[12];
                      spriteData[23] <= memory[11];
                      spriteData[22] <= memory[10];
                      spriteData[21] <= memory[11];
                      spriteData[20] <= memory[10];
                      spriteData[19] <= memory[9];
                      spriteData[18] <= memory[8];
                      spriteData[17] <= memory[9];
                      spriteData[16] <= memory[8];
                      
                      spriteData[15] <= memory[23];
                      spriteData[14] <= memory[22];
                      spriteData[13] <= memory[23];
                      spriteData[12] <= memory[22];
                      spriteData[11] <= memory[21];
                      spriteData[10] <= memory[20];
                      spriteData[9] <= memory[21];
                      spriteData[8] <= memory[20];
                      spriteData[7] <= memory[19];
                      spriteData[6] <= memory[18];
                      spriteData[5] <= memory[19];
                      spriteData[4] <= memory[18];
                      spriteData[3] <= memory[17];
                      spriteData[2] <= memory[16];
                      spriteData[1] <= memory[17];
                      spriteData[0] <= memory[16];
                    end
                  1:
                    begin
                      spriteData[47] <= memory[15];
                      spriteData[46] <= memory[14];
                      spriteData[45] <= memory[15];
                      spriteData[44] <= memory[14];
                      spriteData[43] <= memory[13];
                      spriteData[42] <= memory[12];
                      spriteData[41] <= memory[13];
                      spriteData[40] <= memory[12];
                      spriteData[39] <= memory[11];
                      spriteData[38] <= memory[10];
                      spriteData[37] <= memory[11];
                      spriteData[36] <= memory[10];
                      spriteData[35] <= memory[9];
                      spriteData[34] <= memory[8];
                      spriteData[33] <= memory[9];
                      spriteData[32] <= memory[8];
                      
                      spriteData[31] <= memory[23];
                      spriteData[30] <= memory[22];
                      spriteData[29] <= memory[23];
                      spriteData[28] <= memory[22];
                      spriteData[27] <= memory[21];
                      spriteData[26] <= memory[20];
                      spriteData[25] <= memory[21];
                      spriteData[24] <= memory[20];
                      spriteData[23] <= memory[19];
                      spriteData[22] <= memory[18];
                      spriteData[21] <= memory[19];
                      spriteData[20] <= memory[18];
                      spriteData[19] <= memory[17];
                      spriteData[18] <= memory[16];
                      spriteData[17] <= memory[17];
                      spriteData[16] <= memory[16];
                      
                      spriteData[15] <= memory[31];
                      spriteData[14] <= memory[30];
                      spriteData[13] <= memory[31];
                      spriteData[12] <= memory[30];
                      spriteData[11] <= memory[29];
                      spriteData[10] <= memory[28];
                      spriteData[9] <= memory[29];
                      spriteData[8] <= memory[28];
                      spriteData[7] <= memory[27];
                      spriteData[6] <= memory[26];
                      spriteData[5] <= memory[27];
                      spriteData[4] <= memory[26];
                      spriteData[3] <= memory[25];
                      spriteData[2] <= memory[24];
                      spriteData[1] <= memory[25];
                      spriteData[0] <= memory[24];
                    end
                  2:
                    begin
                      spriteData[47] <= memory[23];
                      spriteData[46] <= memory[22];
                      spriteData[45] <= memory[23];
                      spriteData[44] <= memory[22];
                      spriteData[43] <= memory[21];
                      spriteData[42] <= memory[20];
                      spriteData[41] <= memory[21];
                      spriteData[40] <= memory[20];
                      spriteData[39] <= memory[19];
                      spriteData[38] <= memory[18];
                      spriteData[37] <= memory[19];
                      spriteData[36] <= memory[18];
                      spriteData[35] <= memory[17];
                      spriteData[34] <= memory[16];
                      spriteData[33] <= memory[17];
                      spriteData[32] <= memory[16];
                      
                      spriteData[31] <= memory[31];
                      spriteData[30] <= memory[30];
                      spriteData[29] <= memory[31];
                      spriteData[28] <= memory[30];
                      spriteData[27] <= memory[29];
                      spriteData[26] <= memory[28];
                      spriteData[25] <= memory[29];
                      spriteData[24] <= memory[28];
                      spriteData[23] <= memory[27];
                      spriteData[22] <= memory[26];
                      spriteData[21] <= memory[27];
                      spriteData[20] <= memory[26];
                      spriteData[19] <= memory[25];
                      spriteData[18] <= memory[24];
                      spriteData[17] <= memory[25];
                      spriteData[16] <= memory[24];
                      
                      spriteData[15] <= memory[7];
                      spriteData[14] <= memory[6];
                      spriteData[13] <= memory[7];
                      spriteData[12] <= memory[6];
                      spriteData[11] <= memory[5];
                      spriteData[10] <= memory[4];
                      spriteData[9] <= memory[5];
                      spriteData[8] <= memory[4];
                      spriteData[7] <= memory[3];
                      spriteData[6] <= memory[2];
                      spriteData[5] <= memory[3];
                      spriteData[4] <= memory[2];
                      spriteData[3] <= memory[1];
                      spriteData[2] <= memory[0];
                      spriteData[1] <= memory[1];
                      spriteData[0] <= memory[0];
                    end
                  3:
                    begin
                      spriteData[47] <= memory[31];
                      spriteData[46] <= memory[30];
                      spriteData[45] <= memory[31];
                      spriteData[44] <= memory[30];
                      spriteData[43] <= memory[29];
                      spriteData[42] <= memory[28];
                      spriteData[41] <= memory[29];
                      spriteData[40] <= memory[28];
                      spriteData[39] <= memory[27];
                      spriteData[38] <= memory[26];
                      spriteData[37] <= memory[27];
                      spriteData[36] <= memory[26];
                      spriteData[35] <= memory[25];
                      spriteData[34] <= memory[24];
                      spriteData[33] <= memory[25];
                      spriteData[32] <= memory[24];
                      
                      spriteData[31] <= memory[7];
                      spriteData[30] <= memory[6];
                      spriteData[29] <= memory[7];
                      spriteData[28] <= memory[6];
                      spriteData[27] <= memory[5];
                      spriteData[26] <= memory[4];
                      spriteData[25] <= memory[5];
                      spriteData[24] <= memory[4];
                      spriteData[23] <= memory[3];
                      spriteData[22] <= memory[2];
                      spriteData[21] <= memory[3];
                      spriteData[20] <= memory[2];
                      spriteData[19] <= memory[1];
                      spriteData[18] <= memory[0];
                      spriteData[17] <= memory[1];
                      spriteData[16] <= memory[0];
                      
                      spriteData[15] <= memory[15];
                      spriteData[14] <= memory[14];
                      spriteData[13] <= memory[15];
                      spriteData[12] <= memory[14];
                      spriteData[11] <= memory[13];
                      spriteData[10] <= memory[12];
                      spriteData[9] <= memory[13];
                      spriteData[8] <= memory[12];
                      spriteData[7] <= memory[11];
                      spriteData[6] <= memory[10];
                      spriteData[5] <= memory[11];
                      spriteData[4] <= memory[10];
                      spriteData[3] <= memory[9];
                      spriteData[2] <= memory[8];
                      spriteData[1] <= memory[9];
                      spriteData[0] <= memory[8];
                    end
                endcase
              end
            else //normal color
              begin
                case(byteSelect[1:0])
                  0:
                    begin
                      spriteData[47] <= memory[7];
                      spriteData[46] <= memory[7];
                      spriteData[45] <= memory[6];
                      spriteData[44] <= memory[6];
                      spriteData[43] <= memory[5];
                      spriteData[42] <= memory[5];
                      spriteData[41] <= memory[4];
                      spriteData[40] <= memory[4];
                      spriteData[39] <= memory[3];
                      spriteData[38] <= memory[3];
                      spriteData[37] <= memory[2];
                      spriteData[36] <= memory[2];
                      spriteData[35] <= memory[1];
                      spriteData[34] <= memory[1];
                      spriteData[33] <= memory[0];
                      spriteData[32] <= memory[0];
                      
                      spriteData[31] <= memory[15];
                      spriteData[30] <= memory[15];
                      spriteData[29] <= memory[14];
                      spriteData[28] <= memory[14];
                      spriteData[27] <= memory[13];
                      spriteData[26] <= memory[13];
                      spriteData[25] <= memory[12];
                      spriteData[24] <= memory[12];
                      spriteData[23] <= memory[11];
                      spriteData[22] <= memory[11];
                      spriteData[21] <= memory[10];
                      spriteData[20] <= memory[10];
                      spriteData[19] <= memory[9];
                      spriteData[18] <= memory[9];
                      spriteData[17] <= memory[8];
                      spriteData[16] <= memory[8];
                      
                      spriteData[15] <= memory[23];
                      spriteData[14] <= memory[23];
                      spriteData[13] <= memory[22];
                      spriteData[12] <= memory[22];
                      spriteData[11] <= memory[21];
                      spriteData[10] <= memory[21];
                      spriteData[9] <= memory[20];
                      spriteData[8] <= memory[20];
                      spriteData[7] <= memory[19];
                      spriteData[6] <= memory[19];
                      spriteData[5] <= memory[18];
                      spriteData[4] <= memory[18];
                      spriteData[3] <= memory[17];
                      spriteData[2] <= memory[17];
                      spriteData[1] <= memory[16];
                      spriteData[0] <= memory[16];
                    end
                  1:
                    begin
                      spriteData[47] <= memory[15];
                      spriteData[46] <= memory[15];
                      spriteData[45] <= memory[14];
                      spriteData[44] <= memory[14];
                      spriteData[43] <= memory[13];
                      spriteData[42] <= memory[13];
                      spriteData[41] <= memory[12];
                      spriteData[40] <= memory[12];
                      spriteData[39] <= memory[11];
                      spriteData[38] <= memory[11];
                      spriteData[37] <= memory[10];
                      spriteData[36] <= memory[10];
                      spriteData[35] <= memory[9];
                      spriteData[34] <= memory[9];
                      spriteData[33] <= memory[8];
                      spriteData[32] <= memory[8];
                      
                      spriteData[31] <= memory[23];
                      spriteData[30] <= memory[23];
                      spriteData[29] <= memory[22];
                      spriteData[28] <= memory[22];
                      spriteData[27] <= memory[21];
                      spriteData[26] <= memory[21];
                      spriteData[25] <= memory[20];
                      spriteData[24] <= memory[20];
                      spriteData[23] <= memory[19];
                      spriteData[22] <= memory[19];
                      spriteData[21] <= memory[18];
                      spriteData[20] <= memory[18];
                      spriteData[19] <= memory[17];
                      spriteData[18] <= memory[17];
                      spriteData[17] <= memory[16];
                      spriteData[16] <= memory[16];
                      
                      spriteData[15] <= memory[31];
                      spriteData[14] <= memory[31];
                      spriteData[13] <= memory[30];
                      spriteData[12] <= memory[30];
                      spriteData[11] <= memory[29];
                      spriteData[10] <= memory[29];
                      spriteData[9] <= memory[28];
                      spriteData[8] <= memory[28];
                      spriteData[7] <= memory[27];
                      spriteData[6] <= memory[27];
                      spriteData[5] <= memory[26];
                      spriteData[4] <= memory[26];
                      spriteData[3] <= memory[25];
                      spriteData[2] <= memory[25];
                      spriteData[1] <= memory[24];
                      spriteData[0] <= memory[24];
                    end
                  2:
                    begin
                      spriteData[47] <= memory[23];
                      spriteData[46] <= memory[23];
                      spriteData[45] <= memory[22];
                      spriteData[44] <= memory[22];
                      spriteData[43] <= memory[21];
                      spriteData[42] <= memory[21];
                      spriteData[41] <= memory[20];
                      spriteData[40] <= memory[20];
                      spriteData[39] <= memory[19];
                      spriteData[38] <= memory[19];
                      spriteData[37] <= memory[18];
                      spriteData[36] <= memory[18];
                      spriteData[35] <= memory[17];
                      spriteData[34] <= memory[17];
                      spriteData[33] <= memory[16];
                      spriteData[32] <= memory[16];
                      
                      spriteData[31] <= memory[31];
                      spriteData[30] <= memory[31];
                      spriteData[29] <= memory[30];
                      spriteData[28] <= memory[30];
                      spriteData[27] <= memory[29];
                      spriteData[26] <= memory[29];
                      spriteData[25] <= memory[28];
                      spriteData[24] <= memory[28];
                      spriteData[23] <= memory[27];
                      spriteData[22] <= memory[27];
                      spriteData[21] <= memory[26];
                      spriteData[20] <= memory[26];
                      spriteData[19] <= memory[25];
                      spriteData[18] <= memory[25];
                      spriteData[17] <= memory[24];
                      spriteData[16] <= memory[24];
                      
                      spriteData[15] <= memory[7];
                      spriteData[14] <= memory[7];
                      spriteData[13] <= memory[6];
                      spriteData[12] <= memory[6];
                      spriteData[11] <= memory[5];
                      spriteData[10] <= memory[5];
                      spriteData[9] <= memory[4];
                      spriteData[8] <= memory[4];
                      spriteData[7] <= memory[3];
                      spriteData[6] <= memory[3];
                      spriteData[5] <= memory[2];
                      spriteData[4] <= memory[2];
                      spriteData[3] <= memory[1];
                      spriteData[2] <= memory[1];
                      spriteData[1] <= memory[0];
                      spriteData[0] <= memory[0];
                    end
                  3:
                    begin
                      spriteData[47] <= memory[31];
                      spriteData[46] <= memory[31];
                      spriteData[45] <= memory[30];
                      spriteData[44] <= memory[30];
                      spriteData[43] <= memory[29];
                      spriteData[42] <= memory[29];
                      spriteData[41] <= memory[28];
                      spriteData[40] <= memory[28];
                      spriteData[39] <= memory[27];
                      spriteData[38] <= memory[27];
                      spriteData[37] <= memory[26];
                      spriteData[36] <= memory[26];
                      spriteData[35] <= memory[25];
                      spriteData[34] <= memory[25];
                      spriteData[33] <= memory[24];
                      spriteData[32] <= memory[24];
                      
                      spriteData[31] <= memory[7];
                      spriteData[30] <= memory[7];
                      spriteData[29] <= memory[6];
                      spriteData[28] <= memory[6];
                      spriteData[27] <= memory[5];
                      spriteData[26] <= memory[5];
                      spriteData[25] <= memory[4];
                      spriteData[24] <= memory[4];
                      spriteData[23] <= memory[3];
                      spriteData[22] <= memory[3];
                      spriteData[21] <= memory[2];
                      spriteData[20] <= memory[2];
                      spriteData[19] <= memory[1];
                      spriteData[18] <= memory[1];
                      spriteData[17] <= memory[0];
                      spriteData[16] <= memory[0];
                      
                      spriteData[15] <= memory[15];
                      spriteData[14] <= memory[15];
                      spriteData[13] <= memory[14];
                      spriteData[12] <= memory[14];
                      spriteData[11] <= memory[13];
                      spriteData[10] <= memory[13];
                      spriteData[9] <= memory[12];
                      spriteData[8] <= memory[12];
                      spriteData[7] <= memory[11];
                      spriteData[6] <= memory[11];
                      spriteData[5] <= memory[10];
                      spriteData[4] <= memory[10];
                      spriteData[3] <= memory[9];
                      spriteData[2] <= memory[9];
                      spriteData[1] <= memory[8];
                      spriteData[0] <= memory[8];
                    end
                endcase
              end
          end
        else //not x Expanded
          begin
            spriteData[23:0] <= 24'h000000;
            //only the higher order bits of the buffer used
            case(byteSelect[1:0])
              0:
                begin
                  spriteData[47:40] <= memory[7:0];
                  spriteData[39:32] <= memory[15:8];
                  spriteData[31:24] <= memory[23:16];
                end
              1:
                begin
                  spriteData[47:40] <= memory[15:8];
                  spriteData[39:32] <= memory[23:16];
                  spriteData[31:24] <= memory[31:24];
                end
              2:
                begin
                  spriteData[47:32] <= memory[23:16];
                  spriteData[39:32] <= memory[31:24];
                  spriteData[31:24] <= memory[7:0];
                end
              3:
                begin
                  spriteData[47:40] <= memory[31:24];
                  spriteData[39:32] <= memory[7:0];
                  spriteData[31:24] <= memory[15:8];
                end
            endcase
          end
      end // dmaCount
    else 
      begin
        if(vclk)
          begin
            //this stuff is only done while dma is not active
            if(hCounter == 90)
              begin
                //lock in the sprite data that will be used for the next 2 rasters
                //this should be done after dma is complete...
                rsprite0Data <= sprite0Data;
                rsprite1Data <= sprite1Data;
                rsprite2Data <= sprite2Data;
                rsprite3Data <= sprite3Data;
                rsprite4Data <= sprite4Data;
                rsprite5Data <= sprite5Data;
                rsprite6Data <= sprite6Data;
                rsprite7Data <= sprite7Data;
              end
            else if(hCounter == trigger_dma)
              begin
                //turn off sprite shifting
                shiftSprite0 <= 0;
                shiftSprite1 <= 0;
                shiftSprite2 <= 0;
                shiftSprite3 <= 0;
                shiftSprite4 <= 0;
                shiftSprite5 <= 0;
                shiftSprite6 <= 0;
                shiftSprite7 <= 0;

                if(vCounter[0] | ~r_vDispVisible)
                  begin
                    //update registered video registers
                    //this is done to make sure video changes do not occur
                    //during active rasters since thay have to be duplicated
                    //for vga video - during inactive rasters changes can occur
                    //every raster but still occur off screen...
                    
                    rvic_00 <= vic_00;
                    rvic_01 <= vic_01;
                    rvic_02 <= vic_02;
                    rvic_03 <= vic_03;
                    rvic_04 <= vic_04;
                    rvic_05 <= vic_05;
                    rvic_06 <= vic_06;
                    rvic_07 <= vic_07;
                    rvic_08 <= vic_08;
                    rvic_09 <= vic_09;
                    rvic_0A <= vic_0A;
                    rvic_0B <= vic_0B;
                    rvic_0C <= vic_0C;
                    rvic_0D <= vic_0D;
                    rvic_0E <= vic_0E;
                    rvic_0F <= vic_0F;

                    rvic_11 <= vic_11;

                    rvic_15 <= vic_15;
                    rvic_16 <= vic_16;
                    rvic_17 <= vic_17;
                    rvic_18 <= vic_18;

                    rvic_1B <= vic_1B;
                    rvic_1C <= vic_1C;
                    rvic_1D <= vic_1D;

                    rvic_20 <= vic_20;
                    rvic_21 <= vic_21;
                    rvic_22 <= vic_22;
                    rvic_23 <= vic_23;
                    rvic_24 <= vic_24;
                    rvic_25 <= vic_25;
                    rvic_26 <= vic_26;
                    rvic_27 <= vic_27;
                    rvic_28 <= vic_28;
                    rvic_29 <= vic_29;
                    rvic_2A <= vic_2A;
                    rvic_2B <= vic_2B;
                    rvic_2C <= vic_2C;
                    rvic_2D <= vic_2D;
                    rvic_2E <= vic_2E;
`ifdef mouser
                    rvic_2F <= vic_2F;
`endif
                  end

                if(vCounter[0] & ~dmaEnable)
                  begin
                    dmaCount <= dma_Count;
                    dmaVCounter <= dmaVCounter + H;
                  end
              end
            else if(vCounter < 475)
              begin            
                if(sprite0YIndex[7:0] < (rvic_17[0] ? 8'h2A : 8'h15))
                  begin
                    //I think this is wrong - and all the others below...
                    if(hCounter == rvic_00 - H)
                      begin
                        shiftSprite0 <= H;
                      end
                  end
                  
                if(sprite1YIndex[7:0] < (rvic_17[1] ? 8'h2A : 8'h15))
                  begin
                    if(hCounter == rvic_02 - H)
                      begin
                        shiftSprite1 <= H;
                      end
                  end
                  
                if(sprite2YIndex[7:0] < (rvic_17[2] ? 8'h2A : 8'h15))
                  begin
                    if(hCounter == rvic_04 - H)
                      begin
                        shiftSprite2 <= H;
                      end
                  end
                  
                if(sprite3YIndex[7:0] < (rvic_17[3] ? 8'h2A : 8'h15))
                  begin
                    if(hCounter == rvic_06 - H)
                      begin
                        shiftSprite3 <= H;
                      end
                  end
                  
                if(sprite4YIndex[7:0] < (rvic_17[4] ? 8'h2A : 8'h15))
                  begin
                    if(hCounter == rvic_08 - H)
                      begin
                        shiftSprite4 <= H;
                      end
                  end
                  
                if(sprite5YIndex[7:0] < (rvic_17[5] ? 8'h2A : 8'h15))
                  begin
                    if(hCounter == rvic_0A - H)
                      begin
                        shiftSprite5 <= H;
                      end
                  end
                  
                if(sprite6YIndex[7:0] < (rvic_17[6] ? 8'h2A : 8'h15))
                  begin
                    if(hCounter == rvic_0C - H)
                      begin
                        shiftSprite6 <= H;
                      end
                  end
                  
                if(sprite7YIndex[7:0] < (rvic_17[7] ? 8'h2A : 8'h15))
                  begin
                    if(hCounter == rvic_0E - H)
                      begin
                        shiftSprite7 <= H;
                      end
                  end
              end
              
            if(shiftSprite0)
              begin
                    //I think this is wrong - and all the others below...
                if(rvic_1C[0] & (hCounter[0] != rvic_00[0]))
                  begin
                    rsprite0Data[47:2] <= rsprite0Data[45:0];
                    rsprite0Data[1:0] <= 2'b00;
                  end
                else if(~rvic_1C[0])
                  begin
                    rsprite0Data[47:1] <= rsprite0Data[46:0];
                    rsprite0Data[0] <= L;
                  end
              end
      
            if(shiftSprite1)
              begin
                if(rvic_1C[1] & (hCounter[0] != rvic_02[0]))
                  begin
                    rsprite1Data[47:2] <= rsprite1Data[45:0];
                    rsprite1Data[1:0] <= 2'b00;
                  end
                else if(~rvic_1C[1])
                  begin
                    rsprite1Data[47:1] <= rsprite1Data[46:0];
                    rsprite1Data[0] <= L;
                  end
              end
      
            if(shiftSprite2)
              begin
                if(rvic_1C[2] & (hCounter[0] != rvic_04[0]))
                  begin
                    rsprite2Data[47:2] <= rsprite2Data[45:0];
                    rsprite2Data[1:0] <= 2'b00;
                  end
                else if(~rvic_1C[2])
                  begin
                    rsprite2Data[47:1] <= rsprite2Data[46:0];
                    rsprite2Data[0] <= L;
                  end
              end
      
            if(shiftSprite3)
              begin
                if(rvic_1C[3] & (hCounter[0] != rvic_06[0]))
                  begin
                    rsprite3Data[47:2] <= rsprite3Data[45:0];
                    rsprite3Data[1:0] <= 2'b00;
                  end
                else if(~rvic_1C[3])
                  begin
                    rsprite3Data[47:1] <= rsprite3Data[46:0];
                    rsprite3Data[0] <= L;
                  end
              end
      
            if(shiftSprite4)
              begin
                if(rvic_1C[4] & (hCounter[0] != rvic_08[0]))
                  begin
                    rsprite4Data[47:2] <= rsprite4Data[45:0];
                    rsprite4Data[1:0] <= 2'b00;
                  end
                else if(~rvic_1C[4])
                  begin
                    rsprite4Data[47:1] <= rsprite4Data[46:0];
                    rsprite4Data[0] <= L;
                  end
              end
      
            if(shiftSprite5)
              begin
                if(rvic_1C[5] & (hCounter[0] != rvic_0A[0]))
                  begin
                    rsprite5Data[47:2] <= rsprite5Data[45:0];
                    rsprite5Data[1:0] <= 2'b00;
                  end
                else if(~rvic_1C[5])
                  begin
                    rsprite5Data[47:1] <= rsprite5Data[46:0];
                    rsprite5Data[0] <= L;
                  end
              end
      
            if(shiftSprite6)
              begin
                if(rvic_1C[6] & (hCounter[0] != rvic_0C[0]))
                  begin
                    rsprite6Data[47:2] <= rsprite6Data[45:0];
                    rsprite6Data[1:0] <= 2'b00;
                  end
                else if(~rvic_1C[6])
                  begin
                    rsprite6Data[47:1] <= rsprite6Data[46:0];
                    rsprite6Data[0] <= L;
                  end
              end

            if(shiftSprite7)
              begin
                if(rvic_1C[7] & (hCounter[0] != rvic_0E[0]))
                  begin
                    rsprite7Data[47:2] <= rsprite7Data[45:0];
                    rsprite7Data[1:0] <= 2'b00;
                  end
                else if(~rvic_1C[7])
                  begin
                    rsprite7Data[47:1] <= rsprite7Data[46:0];
                    rsprite7Data[0] <= L;
                  end
              end

            //C64 hCounter must be low enough to get all memory accesses done before it rolls to zero
            //442 is the is the current dma trigger position, 60 memory cycles are needed before it turns zero
            //there is a 1 cycle delay from the time the request is made until the ack
            //shows up, there is a 3 cycle delay from outputing the address
            //to being able to load the data - sprite data takes an extra cycle because of possible expansion.
            if(shiftHCounter != 0)
              begin
                bitmap <= bitmap << 1;
                //rotate the data around for use in a second scan line
                bitmap[0] <= bitmap[319];
        
                shiftHCounter <= shiftHCounter - H;
                delayed_bitmap[2:0] <= { bitmap[319:318], bitmap[0] };
                delayed_color <= color[159:156];
                delayed_char <= char[319:312]; //for color
                
               //shift character and color information every 8th pixel clock
                if(shiftHCounter[0] & ~shiftHCounter[1] & ~shiftHCounter[2])
                  begin
                    //rotate through the 40 character pointers and color ram
                    
                    //these are only used in multicolor character display mode
                    char <= char << 8;
                    char[7:0] <= char[319:312];
                
                    //these are used in every display mode
                    color <= color << 4;
                    color[3:0] <= color[159:156];
                  end
              end
          end //vclk
      end //dma
   end //end always

assign CharLSN = delayed_char[3:0];//char[315:312];
assign CharMSN = delayed_char[7:4];//char[319:316];

assign red[1] = videoRegOut[7];
assign red[0] = videoRegOut[6];
assign green[2] = videoRegOut[5];
assign green[1] = videoRegOut[4];
assign green[0] = videoRegOut[3];
assign blue[2] = videoRegOut[2];
assign blue[1] = videoRegOut[1];
assign blue[0] = videoRegOut[0];

assign spriteMSB[0] = vic_00[8];
assign spriteMSB[1] = vic_02[8];
assign spriteMSB[2] = vic_04[8];
assign spriteMSB[3] = vic_06[8];
assign spriteMSB[4] = vic_08[8];
assign spriteMSB[5] = vic_0A[8];
assign spriteMSB[6] = vic_0C[8];
assign spriteMSB[7] = vic_0E[8];

endmodule

